/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.techfak.scie;

import de.unibi.techfak.scie.analysis.AnalysisAnnotator;
import de.unibi.techfak.scie.annotators.AnnotationCounter;
import de.unibi.techfak.scie.annotators.abbreviation.AbbreviationAnnotator;
import de.unibi.techfak.scie.annotators.animal.ExactAgeAnnotator;
import de.unibi.techfak.scie.annotators.animal.LabGroupAnnotator;
import de.unibi.techfak.scie.annotators.animal.WeightAnnotator;
import de.unibi.techfak.scie.annotators.injury.InjuryAnnotator;
import de.unibi.techfak.scie.annotators.injury.VertebralPositionAnnotator;
import de.unibi.techfak.scie.annotators.injury.VertebralPositionRangeAnnotator;
import de.unibi.techfak.scie.annotators.ontology.OntologyAnnotator;
import de.unibi.techfak.scie.annotators.quantities.DurationAnnotator;
import de.unibi.techfak.scie.annotators.quantities.InexactQuantityAnnotator;
import de.unibi.techfak.scie.annotators.quantities.NumberAnnotator;
import de.unibi.techfak.scie.annotators.quantities.RangeAnnotator;
import de.unibi.techfak.scie.annotators.quantities.TimeAnnotator;
import de.unibi.techfak.scie.annotators.quantities.WordNumberAnnotator;
import de.unibi.techfak.scie.annotators.result.JudgementAnnotator;
import de.unibi.techfak.scie.annotators.result.PValueAnnotator;
import de.unibi.techfak.scie.annotators.result.ResultAnnotator;
import de.unibi.techfak.scie.annotators.structure.HeaderAnnotator;
import de.unibi.techfak.scie.annotators.structure.SectionAnnotator;
import de.unibi.techfak.scie.annotators.structure.StanfordTokenizer;
import de.unibi.techfak.scie.annotators.treatment.DoseAnnotator;
import de.unibi.techfak.scie.annotators.treatment.TreatmentAnnotator;
import de.unibi.techfak.scie.db.DatabaseSingleton;
import de.unibi.techfak.scie.document_import.DocumentImporter;
import de.unibi.techfak.scie.document_import.PDFStructuredTextExtractorWrapper;
import de.unibi.techfak.scie.ontology.DictionaryAlgorithms;
import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.IOException;
import org.apache.uima.UIMAException;
import org.apache.uima.fit.factory.JCasFactory;
import org.apache.uima.fit.util.CasIOUtil;
import org.apache.uima.jcas.JCas;

/**
 * This is the command line interface for the SCIE project. It can be
 * called in two basic modes:
 * <ul><li>Annotation, which takes a .pdf file (or multiple .pdf files) as input
 * and returns the annotations as XCAS output</li>
 * <li>Training, which takes two directories (input and output). It expects
 * tuples of .rel and XCAS files in the input folder and writes the trained
 * annotator models to the output. Optionally you can specify even more details
 * about the training.</li></ul>
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class Main {

	public static final String PROJECT_NAME = "SCIE";
	public static final String TYPESYSTEM = "src/main/resources/de/unibi/techfak/isy/lppba/descriptors/LPPBATypeSystem";

	public static void main(String[] args) {
		if (args.length < 1 || args[0].equals("--help") || args[0].equals("-h")) {
			printHelp();
			return;
		}
		boolean force;
		File input;
		File outputFolder;

		int i = 0;
		force = args[i].equals("--force") || args[i].equals("-f");
		if (force) {
			i++;
		}
		if (args.length < i + 1) {
			throw new UnsupportedOperationException("Expected an input path as " + i + ". argument!");
		}
		input = new File(args[i++]);
		final PDFFilter pdfFilter = new PDFFilter();
		final File[] pdfs;
		if (input.isFile()) {
			if (!pdfFilter.accept(input)) {
				throw new UnsupportedOperationException("The given file is no PDF: " + input.getAbsolutePath());
			}
			pdfs = new File[]{input};
		} else {
			pdfs = input.listFiles(pdfFilter);
			if (pdfs == null || pdfs.length == 0) {
				throw new UnsupportedOperationException("The given directory contained no PDFs: " + input.getAbsolutePath());
			}
		}
		if (args.length < i + 1) {
			outputFolder = new File(".");
			System.out.println("No output directory was given. Using " + outputFolder.getAbsolutePath() + " per default.");
		} else {
			outputFolder = new File(args[i++]);
			if (!outputFolder.isDirectory()) {
				throw new UnsupportedOperationException("The given output path is no directory: " + outputFolder.getAbsolutePath());
			}
			if (!outputFolder.exists()) {
				outputFolder.mkdirs();
				System.out.println("Creating directory: " + outputFolder.getAbsolutePath());
			}
		}
		annotate(pdfs, outputFolder, force);
		System.exit(0);
	}

	private static final class PDFFilter implements FileFilter {

		@Override
		public boolean accept(File file) {
			return file.getName().toLowerCase().endsWith(".pdf");
		}

	}

	private static void printHelp() {
		System.out.println("SCIE -- Spinal Cord Injury Ontology Extraction");
		System.out.println("Copyright (C) 2013, 2014\n");
		System.out.println("Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel");
		System.out.println("");
		System.out.println("Command Line Usage:");
		System.out.println("java -jar " + PROJECT_NAME + ".jar [--force|-f] <input_path> [output_path]");
		System.out.println("where <input_path> either leads to a single PDF file "
				+ "or to a directory containing PDF files");
		System.out.println("Additionally you may set the 'force' flag such that "
				+ "an output file that is already present is overriden.");
		System.out.println("Additionally you may specify an output folder where "
				+ "the output is stored. Per default the current working "
				+ "directory is used.");
		System.out.println("");
		System.out.println("The program will annotate the "
				+ "given PDF files and create an XCAS file containing "
				+ "annotations for each one.");
	}

	public static final String DATABASE_CONFIG_FILE = "config.xml";

	private static void annotate(final File[] pdfs, final File outputFolder, final boolean force) {
		final AnnotatorPipeline pipeline = setUpPipeline();
		DatabaseSingleton.initialize(DATABASE_CONFIG_FILE);
		for (final File pdf : pdfs) {
			System.out.println("Processing file " + pdf.getAbsolutePath());
			try {
				//check the output file
				final File outputFile = new File(outputFolder, pdf.getName().replace(".pdf", ".xml"));
				if (outputFile.exists()) {
					if (!force) {
						throw new IOException("The output file " + outputFile.getAbsolutePath() + " already exists!");
					} else {
						outputFile.delete();
					}
				}
				//import the file.
				JCas jcas = JCasFactory.createJCas(
						"src/main/resources/de/unibi/techfak/isy/lppba/descriptors/LPPBATypeSystem");
				final DocumentImporter instance = new PDFStructuredTextExtractorWrapper(
						new FileInputStream(pdf), jcas);
				instance.doImport();
				//reset the AnnotationCounter
				AnnotationCounter.reset();
				//purge the Words Cache.
				//TODO: This will be unnecessary with the new implementation.
				DictionaryAlgorithms.purgeWordsCache();
				//run the pipeline.
				pipeline.process(jcas);
				//export the result.
				CasIOUtil.writeXCas(jcas, outputFile);
			} catch (UIMAException | IOException ex) {
				System.err.println("The file could not be processed because of an exception during parsing:");
				ex.printStackTrace();
			}
		}
	}

	private static AnnotatorPipeline setUpPipeline() {
		final AnnotatorPipeline out = new AnnotatorPipeline();
		/*
		 * We start with the annotation of document structure. This also
		 * includes a rough linguistic analysis done by the StanfordTokenizer
		 * (POS-tagging).
		 */
		out.getAnnotators().add(new StanfordTokenizer());
		out.getAnnotators().add(new HeaderAnnotator());
		out.getAnnotators().add(new SectionAnnotator());
		/*
		 * Now we annotate ontological instances. Note that this does _not_
		 * depend on the previous annotations. In principle this could be
		 * computed in parallel.
		 */
		out.getAnnotators().add(new OntologyAnnotator());
		/*
		 * We enrich the ontological output by resolving abbreviations.
		 */
		out.getAnnotators().add(new AbbreviationAnnotator());
		/*
		 * Now we add Rule-based annotators. This includes general quantities
		 * and recombinations of other rule-based instances. At times this
		 * depends on the output of the OntologyAnnotator. It also depends
		 * on the output of the StanfordTokenizer.
		 */
		out.getAnnotators().add(new NumberAnnotator());
		out.getAnnotators().add(new WordNumberAnnotator());
		out.getAnnotators().add(new InexactQuantityAnnotator());
		//aggregate numbers to ranges, if possible
		out.getAnnotators().add(new RangeAnnotator());
		//aggregate quantities with units to semantic entities
		out.getAnnotators().add(new TimeAnnotator());
		out.getAnnotators().add(new DurationAnnotator());
		//ANIMAL related quantities
		out.getAnnotators().add(new ExactAgeAnnotator());
		out.getAnnotators().add(new WeightAnnotator());
		//INJURY related quantities
		out.getAnnotators().add(new VertebralPositionAnnotator());
		out.getAnnotators().add(new VertebralPositionRangeAnnotator());
		//TREATMENT related quantities
		out.getAnnotators().add(new DoseAnnotator());
		//RESULT related quantities
		out.getAnnotators().add(new PValueAnnotator());
		//this is not really a quantity
		out.getAnnotators().add(new JudgementAnnotator());
		/*
		 * Now comes the Relation extraction part. We use our heuristic
		 * annotators.
		 */
		out.getAnnotators().add(new LabGroupAnnotator());
		out.getAnnotators().add(new InjuryAnnotator());
		out.getAnnotators().add(new TreatmentAnnotator());
		out.getAnnotators().add(new ResultAnnotator());

		/*
		 * For active learning we also annotate fitting annotations on top.
		 */
		out.getAnnotators().add(new AnalysisAnnotator());
		return out;
	}
}
