/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.annotators.ontology;

import de.unibi.techfak.scie.annotators.AnnotationCounter;
import de.unibi.techfak.scie.db.DatabaseSingleton;
import de.unibi.techfak.isy.lppba.descriptors.ApplicationInstrument;
import de.unibi.techfak.isy.lppba.descriptors.CompoundSupplier;
import de.unibi.techfak.isy.lppba.descriptors.Delivery;
import de.unibi.techfak.isy.lppba.descriptors.Age;
import de.unibi.techfak.isy.lppba.descriptors.Drug;
import de.unibi.techfak.isy.lppba.descriptors.HeaderKeyword;
import de.unibi.techfak.isy.lppba.descriptors.InjuryArea;
import de.unibi.techfak.isy.lppba.descriptors.InjuryDevice;
import de.unibi.techfak.isy.lppba.descriptors.InjuryHeight;
import de.unibi.techfak.isy.lppba.descriptors.InjuryIntesity;
import de.unibi.techfak.isy.lppba.descriptors.InjuryType;
import de.unibi.techfak.isy.lppba.descriptors.OntologyReference;
import de.unibi.techfak.isy.lppba.descriptors.Organism;
import de.unibi.techfak.isy.lppba.descriptors.InvestigationMethods;
import de.unibi.techfak.isy.lppba.descriptors.LanguageQuantifier;
import de.unibi.techfak.isy.lppba.descriptors.Gender;
import de.unibi.techfak.isy.lppba.descriptors.Significance;
import de.unibi.techfak.isy.lppba.descriptors.SpecialLaboratoryAnimal;
import de.unibi.techfak.isy.lppba.descriptors.Trend;
import de.unibi.techfak.isy.lppba.descriptors.WordNumber;
import de.unibi.techfak.scie.ontology.DictionaryMatch;
import de.unibi.techfak.scie.ontology.Node;
import de.unibi.techfak.scie.ontology.OntologyDatabase;
import de.unibi.techfak.scie.ontology.dictionary_name_finder.DictionaryNameFinder;
import de.unibi.techfak.scie.ontology.driver.NodeHandle;
import de.unibi.techfak.scie.ontology.driver.NodeName;
import de.unibi.techfak.scie.ontology.driver.OntologyHandle;
import de.unibi.techfak.scie.probability_model.uima.UIMAWordSyntacticProbabilityModel;
import de.unibi.techfak.scie.util.CachedJCasUtil;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;

import org.apache.uima.analysis_engine.AnalysisEngineProcessException;
import org.apache.uima.jcas.JCas;

/**
 * The OntologyAnnotator is responsible for generating annotations that originate from the ontology
 * database. This class uses the DictionaryNameFinder to find names in the text and refines the
 * result by using prior knowledge about the ontologies.
 *
 * @author Andreas Stoeckel -- astoecke(at)techfak.uni-bielefeld.de
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class OntologyAnnotator extends org.apache.uima.fit.component.JCasAnnotator_ImplBase {

	/**
	 * Sigma used for the gaussian distribution.
	 */
	private static final double SIGMA = 3;
	/**
	 * The minimum probability for which ontology references are created.
	 */
	private static final double MIN_COMPOUND_PROBABILITY = 0.7;
	/**
	 * If we have more than this number of organism annotations in the same region we start
	 * ommitting some of them based on probability.
	 */
	private static final int AMBIGUATION_CUTOFF = 5;
	
	public static class ProbabilityComparator implements
			Comparator<OntologyReference> {
		
		@Override
		public int compare(OntologyReference o1, OntologyReference o2) {
			if (o1 == null) {
				if (o2 == null) {
					return 0;
				}
				return -1;
			} else {
				if (o2 == null) {
					return 1;
				}
			}
			return Double.compare(o1.getProbability(),
					o2.getProbability());
		}
	}

	/**
	 * Creates the OntologyReference instance depending on the ontlogy.
	 */
	public static void createOntologyReferenceAnnotation(JCas jcas,
			DictionaryMatch match) {
		// Fetch some references for convenience
		final NodeHandle nodeHandle = match.getNode();
		final NodeName name = match.getNameDescr();
		final CachedJCasUtil jcasUtil = CachedJCasUtil.getInstance(jcas);

		// Do some additional checks to get rid of unneccessary doubled
		// annotations. First check if there already exists an OntologyReference
		// annotation for the same text.
		HashSet<OntologyReference> otherRefs = new HashSet<>();
		for (final OntologyReference otherRef : jcasUtil.selectCovered(
				OntologyReference.class, match.getBegin(), match.getEnd())) {
			if (otherRef.getOntology().equals(nodeHandle.getOntology().getName())) {
				otherRefs.add(otherRef);
			}
		}
		for (final OntologyReference otherRef : jcasUtil.selectCovering(
				OntologyReference.class, match.getBegin(), match.getEnd())) {
			if (otherRef.getOntology().equals(nodeHandle.getOntology().getName())) {
				otherRefs.add(otherRef);
			}
		}
		// For all those check if the respective ontology node is equal 
		// or a parent or a child node.
		final OntologyDatabase db = DatabaseSingleton.getInstance().
				getOntologyDatabase();
		final Node node = db.getNode(nodeHandle);
		final OntologyHandle ncbiOnto = nodeHandle.getOntology();
		final ArrayList<OntologyReference> remainingRefs = new ArrayList<>();
		for (final OntologyReference otherRef : otherRefs) {
			// If we have two annotations for the same node take the one
			// with maximum probability.
			if (nodeHandle.getId() == otherRef.getId()) {
				if (match.getProbability() < otherRef.getProbability()) {
					return;
				} else {
					otherRef.removeFromIndexes();
					continue;
				}
			}
			// Check if the other node is a parent of this one.
			final NodeHandle otherNode = new NodeHandle(ncbiOnto, (int) otherRef.getId());
			if (node.isA(otherNode)) {
				// If so we only want to store this node, because it is
				// more specific and thus carries more information.
				otherRef.removeFromIndexes();
				continue;
			}
			// Check if this node is a child of the other OntologyReference.		 
			if (db.getNode(otherNode).isA(nodeHandle)) {
				// In that case do not store this node because we only
				// want to store the most specific one.
				return;
			}
			remainingRefs.add(otherRef);
		}

		// Create the instance, depending on the ontology
		OntologyReference ref;
		switch (nodeHandle.getOntology().getName()) {
			case "ncbi":
				ref = new Organism(jcas);
				break;
			case "drug":
				ref = new Drug(jcas);
				Collection<Node> coll=node.getAssociatedNodes();
				((Drug)ref).setIsAnesthesia(false);
				for(Node i:coll){
					if(i.getPrimaryName().toLowerCase().contains("anesthe")){
						((Drug)ref).setIsAnesthesia(true);
					}
				}
				break;
			case "drug2":
				ref = new Drug(jcas);
				Collection<Node> coll2=node.getAssociatedNodes();
				((Drug)ref).setIsAnesthesia(false);
				for(Node i:coll2){
					if(i.getPrimaryName().toLowerCase().contains("anesthe")){
						((Drug)ref).setIsAnesthesia(true);
					}
				}
				break;
			case "Gender":
				ref = new Gender(jcas);
				break;
			case "Significance":
				ref = new Significance(jcas);
				((Significance) ref).setIsSignificant(db.getNode(nodeHandle).
						getPrimaryName().equals("significant"));
				break;
			case "Trend":
				ref = new Trend(jcas);
					((Trend) ref).setNormalizedName(db.getNode(nodeHandle).
						getPrimaryName());
				break;
			case "InjuryType":
				ref = new InjuryType(jcas);
				break;
			case "InjuryDevice":
				ref = new InjuryDevice(jcas);
				break;
			case "InjuryHeight":
				ref = new InjuryHeight(jcas);
				((InjuryHeight) ref).setRegion(db.getNode(nodeHandle).getPrimaryName());
				break;
			case "Age":
				ref = new Age(jcas);
				break;
			case "SpecialLaboratoryAnimal":
				ref = new SpecialLaboratoryAnimal(jcas);
				((SpecialLaboratoryAnimal) ref).setNormalizedName(db.getNode(nodeHandle).
						getPrimaryName());
				break;
			case "Delivery":
				ref = new Delivery(jcas);
				break;
			case "HeaderKeyword":
				ref = new HeaderKeyword(jcas);
				break;
			case "InvestigationMethods":
				ref = new InvestigationMethods(jcas);
				break;
			case "LanguageQuantifier":
				ref = new LanguageQuantifier(jcas);
				((LanguageQuantifier) ref).
						setNormalizedName(db.getNode(nodeHandle).getPrimaryName());
				break;
			case "WordNumber":
				ref = new WordNumber(jcas);
				((WordNumber) ref).setValue(db.getNode(nodeHandle).getId());
				break;
			
			case "ApplicationInstrument":
				ref = new ApplicationInstrument(jcas);
				break;
			case "CompoundSupplier":
				ref = new CompoundSupplier(jcas);
				break;
			case "InjuryArea":
				ref = new InjuryArea(jcas);
				break;
			case "InjuryIntesity":
				ref = new InjuryIntesity(jcas);
				break;
			default:
				ref = new OntologyReference(jcas);
				break;
		}

		// Set all default parameters
		ref.setAnnotationId(AnnotationCounter.getUniqueId());
		ref.setBegin(match.getBegin());
		ref.setEnd(match.getEnd());
		ref.setOntology(nodeHandle.getOntology().getName());
		ref.setId(nodeHandle.getId());
		ref.setNodeName(name.getName());
		ref.setNodeNameType(name.getType() == null ? "" : name.getType());
		ref.setProbability(match.getProbability());

		// now do a last check if there are too many remaining refernces.
		remainingRefs.add(ref);
		//now do a cutoff: If we have more than 5 annotations for the
		//same region we throw away those with low probability
		boolean include = true;
		if (remainingRefs.size() > AMBIGUATION_CUTOFF) {
			Collections.sort(remainingRefs, new ProbabilityComparator());
			for (int i = 0; i < remainingRefs.size() - AMBIGUATION_CUTOFF; i++) {
				final OntologyReference otherRef = remainingRefs.get(i);
				if (otherRef != ref) {
					otherRef.removeFromIndexes();
					jcasUtil.remove(OntologyReference.class, otherRef);
				} else {
					include = false;
				}
			}
		}
		if (include) {
			jcasUtil.add(OntologyReference.class, ref);
			ref.addToIndexes();
		}
	}
	
	@Override
	public void process(JCas jcas) throws AnalysisEngineProcessException {
		// Fetch the reference to the ontology database		
		final OntologyDatabase db = DatabaseSingleton.getInstance().
				getOntologyDatabase();

		// Create a name finder object -- this object is responsible for finding
		// names in the words
		UIMAWordSyntacticProbabilityModel model = new UIMAWordSyntacticProbabilityModel(
				SIGMA, jcas);
		DictionaryNameFinder finder = new DictionaryNameFinder(db,
				jcas.getDocumentText(), MIN_COMPOUND_PROBABILITY, model);

		// Select all words from the text, iterate over them and
		finder.processDocumentText();

		// Let the generator find all word groups
		for (DictionaryMatch nameMatch : finder.calculateMatches()) {
			createOntologyReferenceAnnotation(jcas, nameMatch);
		}
	}
}
