/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.annotators.quantities;

import de.unibi.techfak.scie.annotators.AnnotationCounter;
import java.util.Collection;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.uima.analysis_engine.AnalysisEngineProcessException;
import org.apache.uima.fit.util.JCasUtil;
import org.apache.uima.jcas.JCas;

import de.unibi.techfak.isy.lppba.descriptors.Token;

/**
 * This annotates Roman Numbers like III or LXI.
 *
 * The output of this Annotator are Integer-annotations.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class RomanNumberAnnotator
		extends org.apache.uima.fit.component.JCasAnnotator_ImplBase {

	public RomanNumberAnnotator() {
	}
	public static final Pattern ROMANINT = Pattern.compile("(m{0,3})" +
			"(cm|cd|d?c{0,3})" +
			"(xc|xl|l?x{0,3})" +
			"(ix|iv|v?i{0,3})");

	@Override
	public void process(JCas jcas) throws AnalysisEngineProcessException {
		Collection<Token> words = JCasUtil.select(jcas, Token.class);
		for (final Token word : words) {
			final String wordContent = word.getCoveredText().toLowerCase();
			//parse roman numbers
			final Matcher matcher = ROMANINT.matcher(wordContent);
			if (matcher.matches()) {
				final int val = parseRomanInt(matcher.group(1), matcher.group(2),
						matcher.
						group(3),
						matcher.group(4));
				final de.unibi.techfak.isy.lppba.descriptors.Integer i =
						new de.unibi.techfak.isy.lppba.descriptors.Integer(jcas);
				i.setValue(val);
				i.setBegin(word.getBegin());
				i.setEnd(word.getEnd());
				i.setAnnotationId(AnnotationCounter.getUniqueId());
				i.addToIndexes();
			}
		}
	}

	public int parseRomanInt(final String thousands, final String hundreds,
			final String tens,
			final String ones) {
		int result = 0;
		if (!thousands.isEmpty()) {
			result += thousands.length() * 1000;
		}
		if (!hundreds.isEmpty()) {
			if (hundreds.equals("cm")) {
				result += 900;
			} else if (hundreds.equals("cd")) {
				result += 400;
			} else {
				if (hundreds.startsWith("d")) {
					result += 500;
					result += (hundreds.length() - 1) * 100;
				} else {
					result += hundreds.length() * 100;
				}
			}
		}
		if (!tens.isEmpty()) {
			if (tens.equals("xc")) {
				result += 90;
			} else if (tens.equals("xl")) {
				result += 40;
			} else {
				if (tens.startsWith("l")) {
					result += 50;
					result += (tens.length() - 1) * 10;
				} else {
					result += tens.length() * 10;
				}
			}
		}
		if (!ones.isEmpty()) {
			if (ones.equals("ix")) {
				result += 9;
			} else if (ones.equals("iv")) {
				result += 4;
			} else {
				if (ones.startsWith("v")) {
					result += 5;
					result += (ones.length() - 1);
				} else {
					result += ones.length();
				}
			}
		}

		return result;
	}
}
