/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.annotators.quantities;

import java.util.regex.Pattern;

/**
 * This is an enum describing all Units that are recognized by an UnitQuantityAnnotator.
 *
 * @author Benjamin Paassen - bpaassen(at)techfak.uni-bielefeld.de
 */
public enum Unit {

	//Time units
	MILLISECONDS(new String[]{"ms", "milliseconds?"}, "Milliseconds"),
	SECONDS(new String[]{"s", "seconds?"}, "Seconds"),
	MINUTES(new String[]{"m","min", "minutes?"}, "Minutes"),
	HOURS(new String[]{"h", "hours?"}, "Hours"),
	DAYS(new String[]{"d", "days?"}, "Days"),
	WEEKS(new String[]{"weeks?"}, "Weeks"),
	MONTHS(new String[]{"months?"}, "Months"),
	YEARS(new String[]{"y", "years?"}, "Years"),
	//Weight units
	MILLIGRAMMES(new String[]{"mg", "milligram(?:me)?s?"}, "Milligrammes"),
	GRAMMES(new String[]{"g", "gram(?:me)?s?"}, "Grammes"),
	KILOGRAMMES(new String[]{"kg", "kilogram(?:me)?s?"}, "Kilogrammes"),
	TONNES(new String[]{"t", "ton(?:ne)?s?"}, "Tonnes"),
	IU(new String[]{"iu","units?","u"}, "international unit"),
	POUNDS(new String[]{"lbm?", "℔"}, "Pounds"),
	//Dose units
	IUPERKG(IU,KILOGRAMMES),
	MILLIGRAMMESPERKG(MILLIGRAMMES,KILOGRAMMES),
	GRAMMESPERKG(GRAMMES,KILOGRAMMES),
	
	MILLIGRAMMESPERKGPERHOUR(MILLIGRAMMESPERKG,HOURS),
	GRAMMESPERKGPERHOUR(GRAMMESPERKG,HOURS),;
	
	private final Pattern[] patterns;
	private final String normalizedName;
	//TODO: How to treat multi-token-units like m/s?

	private Unit(String[] regexes, String normalizedName) {
		patterns = new Pattern[regexes.length];
		for (int i = 0; i < regexes.length; i++) {
			patterns[i] = Pattern.compile(regexes[i]);
		}
		this.normalizedName = normalizedName;
	}
	private Unit(Unit enumerator, Unit denominator) {
		patterns= new Pattern[enumerator.getPatterns().length*denominator.getPatterns().length];
		int k=0;
		for (Pattern i:enumerator.getPatterns()) {
			for (Pattern j:denominator.getPatterns()) {
				//TODO: Changed "/" to "\W" because pdf-importer not always read "/" as that character 
				patterns[k] = Pattern.compile(i.toString()+"\\W"+j.toString());
				//patterns[k] = Pattern.compile(i.toString()+"/"+j.toString());
				k++;
			}
		}
		this.normalizedName = enumerator.getNormalizedName()+"/"+denominator.getNormalizedName();
	}

	public Pattern[] getPatterns() {
		return patterns;
	}

	public String getNormalizedName() {
		return normalizedName;
	}
}
