/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.annotators.treatment;

import de.unibi.techfak.scie.annotators.AnnotationCounter;
import java.util.ArrayList;
import java.util.Collection;

import org.apache.uima.analysis_engine.AnalysisEngineProcessException;
import org.apache.uima.fit.util.JCasUtil;
import org.apache.uima.jcas.JCas;
import org.apache.uima.jcas.tcas.Annotation;

import de.unibi.techfak.isy.lppba.descriptors.Delivery;
import de.unibi.techfak.isy.lppba.descriptors.Dose;
import de.unibi.techfak.isy.lppba.descriptors.Drug;
import de.unibi.techfak.isy.lppba.descriptors.Treatment;
import de.unibi.techfak.scie.probability_model.ProbableHypothesis;
import de.unibi.techfak.scie.probability_model.SyntacticProbabilityModel;
import de.unibi.techfak.scie.probability_model.uima.UIMAAnnotationCompositionUtil;
import de.unibi.techfak.scie.probability_model.uima.UIMAAnnotationCompositionUtil.UIMAAnnotationCandidate;
import de.unibi.techfak.scie.probability_model.uima.UIMAAnnotationCompositionUtil.UIMASemanticProbabilityModel;
import de.unibi.techfak.scie.probability_model.uima.UIMACompositor;
import de.unibi.techfak.scie.probability_model.uima.UIMAWordSyntacticProbabilityModel;

public class TreatmentAnnotator extends
		org.apache.uima.fit.component.JCasAnnotator_ImplBase {

	public TreatmentAnnotator() {
	}

	public static final DoseModel DOSEMODEL = new DoseModel();
	public static final DeliveryModel DELIVERYMODEL = new DeliveryModel();
	public static final UIMASemanticProbabilityModel[] MODELS = { DOSEMODEL,
			DELIVERYMODEL };
	public static final Class[] CLASSES = { Dose.class, Delivery.class };
	public static final TreatmentCompositor COMPOSITOR = new TreatmentCompositor();
	/**
	 * This specifies about how far we search for annotations that belong to a
	 * Treatment. For more details on this topic have a look at the
	 * AnnotationCompositionUtil class.
	 */

	public static final double SIGMA = 6;
	/**
	 * Treatment annotations below this probability threshold are thrown away.
	 * This threshold was derived by using statistics over the available 20 test
	 * papers.
	 */
	public static final double THRESHOLD = 0.2;

	@Override
	public void process(JCas jcas) throws AnalysisEngineProcessException {
		final UIMAWordSyntacticProbabilityModel syntacticModel = new UIMAWordSyntacticProbabilityModel(
				SIGMA, jcas);
		COMPOSITOR.setSyntacticModel(syntacticModel);
		final Collection<Drug> drugs = JCasUtil.select(jcas, Drug.class);
		for (final Drug drug : drugs) {
			UIMAAnnotationCompositionUtil.calculateComposita(drug, syntacticModel,
					MODELS, CLASSES, COMPOSITOR, jcas);
		}
	}

	/**
	 * The likelihood model for the Dose.
	 */
	public static class DoseModel extends UIMASemanticProbabilityModel<Dose> {

		public DoseModel() {
		}

		@Override
		public double calculateSemanticProbability(Dose candidate) {
			// every Dose is equally probable.
			return 1;
		}
	}

	/**
	 * The trivial likelihood Model for Delivery.
	 */
	public static class DeliveryModel extends
			UIMASemanticProbabilityModel<Delivery> {

		public DeliveryModel() {
		}

		@Override
		public double calculateSemanticProbability(Delivery candidate) {
			return candidate.getProbability();
		}
	}

	public static class TreatmentCompositor extends UIMACompositor<Treatment> {

		private SyntacticProbabilityModel syntacticModel;

		public TreatmentCompositor() {
		}

		public void setSyntacticModel(SyntacticProbabilityModel syntacticModel) {
			this.syntacticModel = syntacticModel;
		}

		public SyntacticProbabilityModel getSyntacticModel() {
			return syntacticModel;
		}

		@Override
		public Treatment UIMACompose(Annotation core,
				ArrayList<ProbableHypothesis> slotCandidates, int minBegin, int maxEnd,
				JCas jcas) {
			final ProbableHypothesis<UIMAAnnotationCandidate> doseHypo = slotCandidates
					.get(0);
			final ProbableHypothesis<UIMAAnnotationCandidate> admHypo = slotCandidates
					.get(1);
			// calculate the overall probability for this annotation which we
			// arbitrarily define as
			// the average probability of the annotations divided by a normalization
			// factor.
			double probability = 0;
			for (final ProbableHypothesis<UIMAAnnotationCandidate> slotHypo : slotCandidates) {
				if (slotHypo != null) {
					probability += slotHypo.getProbability();
				}
			}
			probability = probability / slotCandidates.size();
			// and we cut off the probability at 1.
			if (probability > 1) {
				probability = 1;
			}
			// if the composition probability is bigger than our threshold, store
			// this result
			if (probability > THRESHOLD) {
				// create annotation
				final Treatment treat = new Treatment(jcas, minBegin, maxEnd);
				treat.setDrug((Drug) core);
				// get all slot fillers
				if (doseHypo != null && doseHypo.getCandidate() != null) {
					treat.setDose((Dose) doseHypo.getCandidate().getAnnotation());
				}
				if (admHypo != null && admHypo.getCandidate() != null) {
					treat.setDelivery((Delivery) admHypo.getCandidate()
							.getAnnotation());
				}
				treat.setProbability(probability);
				treat.setAnnotationId(AnnotationCounter.getUniqueId());
				treat.addToIndexes();
				return treat;
			} else {
				return null;
			}
		}
	}
}