/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.configuration;

import java.io.InputStream;
import java.io.OutputStream;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

/**
 *
 * @author Andreas Stoeckel -- astoecke(at)techfak.uni-bielefeld.de
 */
public class Configuration {

	private final Map<String, Map<String, ConfigurationEntry>> configurationSets;

	private final Map<String, Map<String, String>> valueSets;

	public Configuration() {
		this.configurationSets = new HashMap();
		this.valueSets = new HashMap();
	}

	public void registerConfigurationSet(String name,
			Collection<ConfigurationEntry> entries) {
		// Make sure the configuration set does not exist yet.
		name = name.toLowerCase();
		if (configurationSets.containsKey(name)) {
			throw new RuntimeException("The configuration set with the name \""
					+ name + "\" was already registered!");
		}

		// Create a new hash map and store it in the configurationSets structure.
		Map<String, ConfigurationEntry> entryMap = new HashMap();
		this.configurationSets.put(name, entryMap);

		// Store each configuration entry in the entryMap
		for (ConfigurationEntry entry : entries) {
			entryMap.put(entry.getKey(), entry);
		}
	}

	public void registerConfigurationSet(String name, ConfigurationEntry[] entries) {
		registerConfigurationSet(name, Arrays.asList(entries));
	}

	private ConfigurationEntry getConfigurationEntry(String set, String key) {
		// Return the configuration entry element
		Map<String, ConfigurationEntry> entryMap = configurationSets.get(set);
		if (entryMap != null) {
			return entryMap.get(key);
		}
		return null;
	}

	public String getValue(String set, String key) {
		// Convert the given strings to lowercase letters
		set = set.toLowerCase();
		key = key.toLowerCase();

		// Make sure the configuration entry exists
		ConfigurationEntry entry = getConfigurationEntry(set, key);
		if (entry != null) {
			// Check whether a value was stored for this configuration entry,
			// if not, return the default value
			Map<String, String> valueMap = valueSets.get(set);
			if (valueMap != null && valueMap.containsKey(key)) {
				return valueMap.get(key);
			} else {
				return entry.getDefault();
			}
		}
		return null;
	}

	public boolean setValue(String set, String key, String value) throws ValidityException {
		// Convert the given strings to lowercase letters
		set = set.toLowerCase();
		key = key.toLowerCase();

		// Make sure the configuration entry exists
		ConfigurationEntry entry = getConfigurationEntry(set, key);
		if (entry != null) {
			// Make sure the given value is valid
			if (!entry.isValid(value)) {
				throw new ValidityException(key, value, "Configuration set: " + set);
			}

			// Create the value map entry if it does not exist yet.
			Map<String, String> valueMap = valueSets.get(set);
			if (valueMap == null) {
				valueMap = new HashMap();
				valueSets.put(set, valueMap);
			}

			// Store the value in the value map
			valueMap.put(key, value);
			return true;
		}
		return false;
	}

	public ConfigurationSet getSet(String set) {
		// TODO Make this better
		//if (configurationSets.containsKey(set)) {
			return new ConfigurationSet(set, this);
		//} 
		//return null;
	}

	public void reset() {
		this.configurationSets.clear();
		this.valueSets.clear();
	}

	private enum XMLParserState {

		stateTopLevel, stateRootNode, stateSet;
	};

	private void parseXMLStream(XMLStreamReader parser) throws XMLStreamException, ValidityException {
		String currentSet = null;
		XMLParserState state = XMLParserState.stateTopLevel;

		while (parser.hasNext()) {
			int eventType = parser.getEventType();
			String tagName = "";
			if (eventType == XMLStreamConstants.START_ELEMENT
					|| eventType == XMLStreamConstants.END_ELEMENT) {
				tagName = parser.getLocalName();
			}

			switch (eventType) {
				case XMLStreamConstants.END_DOCUMENT:
					parser.close();
					break;

				case XMLStreamConstants.START_ELEMENT:
					switch (state) {
						case stateTopLevel:
							if ("configuration".equals(tagName)) {
								state = XMLParserState.stateRootNode;
							}
							break;

						case stateRootNode:
							currentSet = tagName;
							state = XMLParserState.stateSet;
							break;

						case stateSet:
							final String key = tagName;
							final String value = parser.getElementText();
							setValue(currentSet, key, value);
							break;
					}
					break;

				case XMLStreamConstants.END_ELEMENT:
					switch (state) {
						case stateRootNode:
							if ("configuration".equals(tagName)) {
								state = XMLParserState.stateTopLevel;
							}
							break;

						case stateSet:
							if (currentSet.equals(tagName)) {
								currentSet = null;
								state = XMLParserState.stateRootNode;
							}
							break;
					}

					break;
			}

			// Only continue if the event type did not change over the course of
			// the processing -- otherwise iterate again in order to process the
			// new event
			if (parser.getEventType() == eventType) {
				parser.next();
			}
		}
	}

	private void writeXMLStream(XMLStreamWriter writer) throws XMLStreamException {
		writer.writeStartDocument();
		writer.writeStartElement("configuration");
		for (String set : valueSets.keySet()) {
			writer.writeStartElement(set);
			Map<String, String> valueMap = valueSets.get(set);
			for (String key : valueMap.keySet()) {
				writer.writeStartElement(key);
				writer.writeCharacters(valueMap.get(key));
				writer.writeEndElement();
			}
			writer.writeEndElement();
		}
		writer.writeEndElement();
		writer.writeEndDocument();
		writer.flush();
	}

	public void loadXMLConfiguration(InputStream is) throws ValidityException {
		try {
			XMLInputFactory factory = XMLInputFactory.newInstance();
			parseXMLStream(factory.createXMLStreamReader(is));
		} catch (XMLStreamException ex) {
			throw new ValidityException("Error while parsing XML stream: "
					+ ex.getMessage());
		}
	}

	public void storeXMLConfiguration(OutputStream os) {
		try {
			XMLOutputFactory factory = XMLOutputFactory.newInstance();
			writeXMLStream(factory.createXMLStreamWriter(os));
		} catch (XMLStreamException ex) {
			throw new RuntimeException("Error while writing XML stream");
		}
	}

	/**
	 * Calculates a hash that represents the current configuration. This
	 * hash chages as soon as a single key or value in the configuration chages.
	 *
	 * @return A hash that represents the current configuration file content.
	 */
	public String configurationHash() {
		// Create the SHA hash generator
		MessageDigest dig;
		try {
			dig = MessageDigest.getInstance("SHA-1");
		} catch (NoSuchAlgorithmException ex) {
			throw new RuntimeException(ex);
		}

		// Iterate over all value sets and all key-value pairs
		for (String setKey : valueSets.keySet()) {
			// Iterate over all set entries, skip empty ones
			Map<String, String> setEntries = valueSets.get(setKey);
			if (setEntries.isEmpty()) {
				continue;
			}

			// Store the set key in the hash
			dig.update(setKey.getBytes());

			for (String key : setEntries.keySet()) {
				dig.update(key.getBytes());
				dig.update(setEntries.get(key).getBytes());
			}
		}

		// Convert the digest to a hex string
		byte buf[] = dig.digest();
		StringBuilder hexString = new StringBuilder();
		for (byte c: buf) {
			hexString.append(Integer.toHexString(0xFF & c));
		}
		return hexString.toString();
	}

}
