/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.ontology;

import de.unibi.techfak.scie.ontology.driver.OntologyDatabaseDriver;
import de.unibi.techfak.scie.ontology.driver.NodeDescriptor;
import de.unibi.techfak.scie.ontology.driver.NodeHandle;
import de.unibi.techfak.scie.ontology.driver.OntologyHandle;
import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Timer;
import java.util.TimerTask;

/**
 * The NodeFactory class is used to convert NodeHandles and NodeDescriptors to
 * the easier to use Node instances -- the translation from NodeHandle to Node
 * is cached in order to always return the same node instance and thus achieve
 * consistency throughout the application.
 *
 * @author Andreas Stoeckel -- astoecke(at)techfak.uni-bielefeld.de
 */
final class NodeFactory {

	private class CacheCleanupTask extends TimerTask {

		/**
		 * Time between cache cleanup runs.
		 */
		private final int CACHE_CLEANUP_PERIOD = 10000;

		/**
		 * Timer instance used for periodical execution of the task.
		 */
		private final Timer timer;

		/**
		 * Reference to the cache that should be cleared.
		 */
		private final Map<NodeHandle, WeakReference<Node>> cache;

		/**
		 * Reference to the lock object which is used for synchronization.
		 */
		private final Object cacheLock;

		/**
		 * Creates the CacheCleanupTask.
		 *
		 * @param cache is a reference to the cache that should be cleared.
		 */
		public CacheCleanupTask(Map<NodeHandle, WeakReference<Node>> cache,
				Object cacheLock) {
			this.timer = new Timer();
			this.cache = cache;
			this.cacheLock = cacheLock;
		}

		/**
		 * Starts the periodical cleanup.
		 */
		public void start() {
			timer.scheduleAtFixedRate(this, CACHE_CLEANUP_PERIOD,
					CACHE_CLEANUP_PERIOD);
		}

		/**
		 * Deletes all handles from the cache which have a null reference.
		 */
		@Override
		public void run() {
			synchronized (cacheLock) {
				// Search all nodes which have no valid node instance attached
				// to them.
				Collection<NodeHandle> nodeHandles = new ArrayList();
				for (NodeHandle node : cache.keySet()) {
					if (cache.get(node).get() == null) {
						nodeHandles.add(node);
					}
				}

				// Remove those nodes from the cache
				for (NodeHandle node : nodeHandles) {
					cache.remove(node);
				}
			}
		}

	};

	/**
	 * Reference to the OntologyDatabaseDriver.
	 */
	private final OntologyDatabaseDriver driver;

	/**
	 * Object used for synchronization with the cache cleanup task.
	 */
	private final Object cacheLock = new Object();

	/**
	 * Map between node handles and the corresponding Node instances. Access is
	 * synchronized using the "cacheLock".
	 */
	private final Map<NodeHandle, WeakReference<Node>> cache;

	/**
	 * Task which periodically purges invalid references from the cache.
	 */
	private final CacheCleanupTask cacheCleanupTask;

	/**
	 * Creates an instance of the NodeFactory class.
	 *
	 * @param driver is a reference to the OntologyDatabaseDriver.
	 */
	public NodeFactory(OntologyDatabaseDriver driver) {
		this.driver = driver;
		this.cache = new HashMap();
		this.cacheCleanupTask = new CacheCleanupTask(cache, cacheLock);
		this.cacheCleanupTask.start();
	}

	/**
	 * Makes sure all threads used by the node factory are properly shut down.
	 */
	public void close() {
		cacheCleanupTask.timer.cancel();
		cacheCleanupTask.timer.purge();
	}

	/**
	 * Returns the reference to the database driver.
	 *
	 * @return the reference to the database driver.
	 */
	public OntologyDatabaseDriver getDriver() {
		return driver;
	}

	/**
	 * Registers a node descriptor in the cache -- if the given instance already
	 * is an instance of the Node class, it is simply added to the map (if there
	 * is no entry yet), otherwisew a node instance is created from the given
	 * descriptor and stored in the cache.
	 *
	 * @param descr is the NodeDescriptor instance that should be stored in the
	 * cache.
	 */
	public void registerNode(NodeDescriptor descr) {
		synchronized (cacheLock) {
			// Do not overwrite cache entries.
			if (!cache.containsKey(descr.getHandle())) {
				// Convert the given node to a Node if it isn't already
				Node node;
				if (!(descr instanceof Node)) {
					node = toNode(descr);
				} else {
					node = (Node) descr;
				}

				// Register the node within the cache if it doesn't already exist
				cache.put(node.getHandle(), new WeakReference(node));
			}
		}
	}

	/**
	 * Removes the given node from the cache -- it is important to call this
	 * method if the given node has been deleted.
	 *
	 * @param handle is the node that should be removed from the cache.
	 */
	public void purgeNode(NodeHandle handle) {
		synchronized (cacheLock) {
			cache.remove(handle);
		}
	}

	/**
	 * Purges the nodes of an entire ontology from the cache.
	 *
	 * @param handle is the ontology for which all nodes in the cache should be
	 * removed.
	 */
	public void purgeOntology(OntologyHandle handle) {
		synchronized (cacheLock) {
			// Search all nodes which belong to the given ontology
			Collection<NodeHandle> nodeHandles = new ArrayList();
			for (NodeHandle node : cache.keySet()) {
				if (node.getOntology().equals(handle)) {
					nodeHandles.add(node);
				}
			}

			// Remove those nodes from the cache
			for (NodeHandle node : nodeHandles) {
				cache.remove(node);
			}
		}
	}

	/**
	 * Converts the given node handle to a Node instance. Fetches the node
	 * descriptor from the database if the node is not yet in the cache,
	 * otherwise the Node instance from the cache is returned.
	 *
	 * @param handle is the node handle that should be looked up.
	 * @param descr may contain the node descriptor of the given handle if it is
	 * known to the caller. Otherwise null may be passed.
	 * @param cacheOnly if true, no database lookups are made.
	 * @return the Node instance corresponding to the given handle or null if a)
	 * the node is not in the cache and cacheOnly is true b) the node is not in
	 * the cache and the database returns null for the given handle.
	 */
	public Node toNode(NodeHandle handle, NodeDescriptor descr, boolean cacheOnly) {
		// Pass null through
		if (handle == null) {
			return null;
		}

		// Check whether a node with the given handle is already cached
		Node res = null;
		WeakReference<Node> ref;
		synchronized (cacheLock) {
			ref = cache.get(handle);
		}
		if (ref != null) {
			res = ref.get();
		}

		// Read the node from the database if it was not cached and we're allows
		// to do so (either the descriptor is given or cacheOnly is false)
		if (res == null && !(cacheOnly && descr == null)) {
			if (descr == null) {
				descr = driver.getNode(handle);

				// Abort if the node does not exist
				if (descr == null) {
					return null;
				}
			}
			res = new Node(this, descr);
			synchronized (cacheLock) {
				cache.put(handle, new WeakReference(res));
			}
		}
		return res;
	}

	/**
	 * Converts the given node handle to a Node instance. Fetches the node
	 * descriptor from the database if the node is not yet in the cache,
	 * otherwise the Node instance from the cache is returned.
	 *
	 * @param handle is the node handle that should be looked up.
	 * @param cacheOnly if true, no database lookups are made.
	 * @return the Node instance corresponding to the given handle or null if a)
	 * the node is not in the cache and cacheOnly is true b) the node is not in
	 * the cache and the database returns null for the given handle.
	 */
	public Node toNode(NodeHandle handle, boolean cacheOnly) {
		return toNode(handle, null, cacheOnly);
	}

	/**
	 * Converts the given node handle to a Node instance. Fetches the node
	 * descriptor from the database if the node is not yet in the cache,
	 * otherwise the Node instance from the cache is returned.
	 *
	 * @param handle is the node handle that should be looked up.
	 * @return the Node instance corresponding to the given handle or null if
	 * the node is not in the cache and the database returns null for the given
	 * handle.
	 */
	public Node toNode(NodeHandle handle) {
		return toNode(handle, null, false);
	}

	/**
	 * Converts the given node descriptor to a Node instance. If the node handle
	 * of the descriptor is already in the cache, this instance is returned
	 * (Note: this may cause data that has been stored in the NodeDescriptor to
	 * be lost!), if the node descriptor is an instance of Node and the node
	 * handle is not in the cache
	 *
	 * @param descr is the node handle that should be looked up.
	 * @return the Node instance corresponding to the given handle or null if
	 * the node is not in the cache and the database returns null for the given
	 * handle.
	 */
	public Node toNode(NodeDescriptor descr) {
		// Do nothing if the given descriptor already is a node
		if (descr instanceof Node) {
			registerNode(descr);
			return (Node) descr;
		}

		// Pass null through
		if (descr == null) {
			return null;
		}

		// Call the toNode function and pass the descriptor itself as the
		// descriptor that should be used.
		return toNode(descr.getHandle(), descr, true);
	}

	/**
	 * Efficiently converts all given node handles to Node instances by
	 * boundeling database lookups into a single driver call.
	 *
	 * @param handles is a list of handles that should be converted to instances
	 * of Node.
	 * @return the list of converted node handles.
	 */
	public List<Node> handlesToNodeList(Collection<NodeHandle> handles) {
		// If "null" is given, pass it through
		if (handles == null) {
			return null;
		}

		// First translate all given handles using the cache -- all handles that
		// cannot be resolved are fetched from the database.
		ArrayList<Node> res = new ArrayList(handles.size());
		Map<NodeHandle, Integer> resolve = new HashMap();

		// Resolve all elements from the cache -- remember the elements that
		// could not be resolved.
		int i = 0;
		for (NodeHandle handle : handles) {
			if (handle != null) {
				Node node = toNode(handle, true);
				res.add(node);
				if (node == null) {
					resolve.put(handle, i);
				}
				i++;
			} else {
				res.add(null);
			}
		}

		// Resolve all remaining elements, add these to the cache
		List<NodeDescriptor> descriptors = driver.getNodes(resolve.keySet());
		for (NodeDescriptor descriptor : descriptors) {
			res.set(resolve.get(descriptor.getHandle()), toNode(descriptor));
		}

		return res;
	}

	/**
	 * Converts all given node descriptors to Node instances.
	 *
	 * @param handles is a list of descriptors that should be converted to
	 * instances of Node.
	 * @return the list of converted node descriptors.
	 */
	public List<Node> descriptorsToNodeList(Collection<NodeDescriptor> descriptors) {
		// If "null" is given, pass it through
		if (descriptors == null) {
			return null;
		}

		ArrayList<Node> res = new ArrayList(descriptors.size());
		for (NodeDescriptor descriptor : descriptors) {
			res.add(toNode(descriptor));
		}
		return res;
	}

}
