/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.ontology.dictionary_name_finder;

import de.unibi.techfak.scie.ontology.DictionaryAlgorithms;
import de.unibi.techfak.scie.ontology.DictionaryAlgorithms.WordPosition;
import de.unibi.techfak.scie.ontology.DictionaryMatch;
import de.unibi.techfak.scie.ontology.driver.ComparisonMode;
import de.unibi.techfak.scie.ontology.driver.NodeHandle;
import de.unibi.techfak.scie.ontology.driver.NodeName;
import de.unibi.techfak.scie.probability_model.AnnotationCandidate;
import de.unibi.techfak.scie.probability_model.AnnotationCompositionUtil;
import de.unibi.techfak.scie.probability_model.AnnotationCompositor;
import de.unibi.techfak.scie.probability_model.SemanticProbabilityModel;
import de.unibi.techfak.scie.probability_model.ProbableHypothesis;
import de.unibi.techfak.scie.probability_model.SyntacticProbabilityModel;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;

/**
 * The OntlogyWordGroup class represents a list of OntologyWordReference
 * instances and their total probability.
 *
 * @author Andreas Stoeckel -- astoecke(at)techfak.uni-bielefeld.de
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
class WordGroup {

	private static class WordSemanticModel implements
			SemanticProbabilityModel<WordReference> {

		private final int maxCharCount;

		public WordSemanticModel(int maxCharCount) {
			this.maxCharCount = maxCharCount;
		}

		@Override
		public double calculateSemanticProbability(WordReference candidate) {
			return candidate.getMatch().getRelativeMatchProbability(maxCharCount);
		}
	}

	/**
	 * The word instances are stored hashed by their position in the text. The
	 * hash is calculated by dividing by a certain value. This way the order of
	 * the instances is preserved and a fast lookup of "relevant" word instances
	 * relative to a certain word can be created.
	 */
	private static final int WORD_INSTANCE_HASH_DIVISOR = 100;

	/**
	 * The maximum distance (in characters) for which two word instances are
	 * even considered as corresponding to each other.
	 */
	private static final int MAX_WORD_INSTANCE_DISTANCE = 1000;

	/**
	 * Word begin to index map.
	 */
	private final Map<Integer, Integer> wordStartMap;

	/**
	 * The position of the first whitespace in the name.
	 */
	private final int firstWhitespaceIdx;

	/**
	 * Contains all word references that belong to this WordGroup. Maps from the
	 * word index to a collection of corresponding word references.
	 */
	private final List<Map<Integer, Collection<WordReference>>> wordInstances;

	/**
	 * Reference to the node this word group refers to.
	 */
	private final NodeHandle node;

	/**
	 * Reference to the node name this word group refers to.
	 */
	private final NodeName name;

	/**
	 * Current comparison mode.
	 */
	private final ComparisonMode mode;

	/**
	 * Specifies whether the comparison mode is either LOWERCASE or STRICT.
	 */
	private final boolean isStrict;

	/**
	 * Maximum character count that can be matched.
	 */
	private int maxCharCount;

	/**
	 * Semantic word probability model.
	 */
	private final WordSemanticModel wordSemanticModel;

	/**
	 * Initializes the word group.
	 *
	 * @param node is a reference to the node handle this group belongs to.
	 * @param name is the node name this word group belongs to.
	 */
	public WordGroup(NodeName name, NodeHandle node) {
		final String s = name.getName();
		this.node = node;
		this.name = name;
		this.wordStartMap = calculateWordStartMap(s);
		this.firstWhitespaceIdx = DictionaryAlgorithms.firstWhitespaceIndex(s);
		this.mode = name.getComparisonMode();
		this.isStrict = mode.getValue() >= ComparisonMode.LOWERCASE.getValue();

		// Initialize the wordInstances list
		final int wordCount = wordStartMap.size();
		wordInstances = new ArrayList(wordCount);
		for (int i = 0; i < wordCount; i++) {
			wordInstances.add(new HashMap());
		}

		// Create the word semantic model
		this.wordSemanticModel = new WordSemanticModel(maxCharCount);
	}

	/**
	 * Calculates the hash value for a given word reference.
	 *
	 * @param ref is the reference for which the hash value should be
	 * calculated.
	 * @return the hash value.
	 */
	private int calculateWordReferenceHash(WordReference ref) {
		return ref.getBegin() / WORD_INSTANCE_HASH_DIVISOR;
	}

	/**
	 * Returns all WordReferences for a given word index that are relevant for
	 * the given center word reference. The count of "relevant" words is
	 * normally significantly smaller than the total count of word instances for
	 * this word.
	 *
	 * @param center is the center WordReference for which all relevant word
	 * instances should be returned.
	 * @param wordIdx is the word instance index.
	 * @return a collection of relevant word instances.
	 */
	private Collection<WordReference> getRelevantWordInstances(WordReference center, int wordIdx) {
		Collection<WordReference> res = new LinkedList();
		Map<Integer, Collection<WordReference>> idxMap = wordInstances.get(wordIdx);
		if (idxMap != null) {
			final int minPos = center.getBegin() - MAX_WORD_INSTANCE_DISTANCE;
			final int maxPos = center.getEnd() + MAX_WORD_INSTANCE_DISTANCE;
			final int minHash = Math.max(0, (int) Math.floor((double) minPos / (double) WORD_INSTANCE_HASH_DIVISOR));
			final int maxHash = (int) Math.ceil((double) maxPos / (double) WORD_INSTANCE_HASH_DIVISOR);
			for (int i = minHash; i < maxHash; i++) {
				Collection<WordReference> refs = idxMap.get(i);
				if (refs != null) {
					res.addAll(refs);
				}
			}
		}
		return res;
	}

	/**
	 * Returns all WordReferences for a given word index.
	 *
	 * @param wordIdx is the word instance index.
	 * @return a collection of word instances.
	 */
	private Collection<WordReference> getAllWordInstances(int wordIdx) {
		Collection<WordReference> res = new LinkedList();
		Map<Integer, Collection<WordReference>> idxMap = wordInstances.get(wordIdx);
		if (idxMap != null) {
			for (Collection<WordReference> refs : idxMap.values()) {
				res.addAll(refs);
			}
		}
		return res;
	}

	/**
	 * Creates a mapping between possible word beginnings and an arbitary index
	 * of this word.
	 */
	private Map<Integer, Integer> calculateWordStartMap(String s) {
		// Collect all unabigous word start positions
		maxCharCount = 0;
		SortedSet<Integer> set = new TreeSet();
		for (WordPosition word : DictionaryAlgorithms.getWords(s, isStrict)) {
			maxCharCount += word.getEnd() - word.getBegin();
			set.add(word.getBegin());
		}

		// Create the result map, which maps from an begin index to a bucket
		// index in the wordInstances map
		int i = 0;
		Map<Integer, Integer> res = new TreeMap();
		for (int begin : set) {
			res.put(begin, i++);
		}
		return res;
	}

	/**
	 * Adds a new word reference.
	 *
	 * @param ref is the OntlogyWordReference instance that should be added to
	 * the WordGroup.
	 */
	public void addWordReference(WordReference ref) {
		// Fetch the word index and insert the word reference into the correct
		// position in the wordInstances map
		final Integer wordIdx = wordStartMap.get(ref.getMatch().getBegin());
		if (wordIdx != null) {
			// Calculate the hash value for the given word reference
			int hashValue = calculateWordReferenceHash(ref);

			// Check whether the bucket for this word reference already exists,
			// if not, create it
			Map<Integer, Collection<WordReference>> idxMap = wordInstances.get(wordIdx);
			Collection<WordReference> refs = idxMap.get(hashValue);
			if (refs == null) {
				refs = new LinkedList();
				idxMap.put(hashValue, refs);
			}
			refs.add(ref);
		}
	}

	public Collection<DictionaryMatch> deriveReferences(String documentText,
			double pThresh, SyntacticProbabilityModel model) {
		List<DictionaryMatch> res;
		if (name.getComparisonMode() == ComparisonMode.FUZZY) {
			res = deriveFuzzyReferences(pThresh, model);
		} else {
			res = deriveStrictReferences(documentText);
		}
		return DictionaryMatch.reduceMatches(documentText, res);
	}

	/**
	 * Derives fuzzy references for the current word group by using a
	 * probability model.
	 *
	 * @param pThresh is the minimum a-posteriori probability for which an
	 * annotation is generated.
	 */
	private List<DictionaryMatch> deriveFuzzyReferences(double pThresh,
			SyntacticProbabilityModel model) {
		final DictionaryMatchCompositor compositor
				= new DictionaryMatchCompositor(name, node, pThresh, maxCharCount);
		List<DictionaryMatch> res = new ArrayList();

		for (int i = 0; i < wordInstances.size(); i++) {
			for (WordReference center : getAllWordInstances(i)) {
				// For each other word in the word group, calculate all possible
				// combinations of them to compose a WordGroup.

				// Retrieve the candidates.
				final Collection[] candidates = new Collection[wordInstances.
						size() - 1];
				final SemanticProbabilityModel[] modelArr
						= new SemanticProbabilityModel[wordInstances.size() - 1];
				for (int j = 0; j < wordInstances.size(); j++) {
					// Skip the current center word
					if (i == j) {
						continue;
					}

					final int idx = (j < i) ? j : j - 1;
					modelArr[idx] = wordSemanticModel;

					// Only add the word instances that do not overlap with the
					// current center word
					candidates[idx] = getRelevantWordInstances(center, j);
				}

				// Calculate the matches.
				res.addAll(AnnotationCompositionUtil.calculateComposita(center,
						model, modelArr, candidates, compositor));
			}
		}
		return res;
	}

	/**
	 * Calculates the minimum character count for a match given the given string
	 * and comparison mode.
	 */
	private static int calcCharacterCount(String n, ComparisonMode mode) {
		final String stripped;
		switch (mode) {
			case SOFT:
				stripped = DictionaryAlgorithms.stripNonAlphabetic(n);
				break;
			case CLEAN:
				stripped = DictionaryAlgorithms.stripNonAlphanumeric(n);
				break;
			case LOWERCASE:
			case STRICT:
				stripped = DictionaryAlgorithms.stripWhitespace(n);
				break;
			default:
				throw new AssertionError("Unexpected comparison mode "
						+ mode.name());
		}
		return stripped.length();
	}

	/**
	 * Derives possible strict references from the word group. Strict references
	 * are characterized by maintining the order of the text and not allowing
	 * non-alphanumeric characters between the words and matching at least as
	 * many alphanumeric characters as present in the word.
	 *
	 * @param documentText is the underlying document text. It is used to decide
	 * whether two words are only seperated by whitspace/non-alphanumeric
	 * characters.
	 */
	private List<DictionaryMatch> deriveStrictReferences(
			String documentText) {
		// Create the result list, count how many characters should be found
		List<DictionaryMatch> res = new ArrayList();
		final String n = name.getName();
		final int minCharacterCount = calcCharacterCount(n, mode);

		// Iterate over all word start positions
		for (int i = 0; i < wordInstances.size(); i++) {
			for (WordReference wI : getAllWordInstances(i)) {
				// Early abortion: We cannot hope to find a strict match starting
				// with another word instance i if we are beyond the first word
				if (i > 0 && wI.getMatch().getBegin() > firstWhitespaceIdx) {
					return res;
				}

				final int wIB = wI.getBegin();
				final int wIE = wI.getEnd();
				final String wIT = documentText.substring(wIB, wIE);
				int currentEnd = wI.getEnd(); // The current end position of wI
				int currentCharacterCount = calcCharacterCount(wIT, mode);

				for (int j = i + 1; j < wordInstances.size(); j++) {
					int minEnd = Integer.MAX_VALUE;
					int minEndCharacterDelta = 0;
					for (WordReference wJ : getRelevantWordInstances(wI, j)) {
						final int wJB = wJ.getBegin();
						final int wJE = wJ.getEnd();
						final String wJT = documentText.substring(wJB, wJE);

						// Skip all word references where the current end
						// position is larger than the begin position of wJ.
						if (currentEnd >= wJ.getBegin()) {
							continue;
						}

						// Check whether the whitespace between wI and wJ
						// contains alphanumeric characters -- if yes, continue
						final String gapText = documentText.substring(currentEnd, wJB);
						if (DictionaryAlgorithms.containsAlphanumericChar(gapText)) {
							continue;
						}

						// Handle the LOWERCASE and STRICT modes here only whitespace
						// may occur between the words. And there has to be more than
						// one whitespace character!
						if (isStrict && (gapText.length() < 1
								|| DictionaryAlgorithms.stripWhitespace(gapText).length() > 0)) {
							continue;
						}

						// Advance the "currentEnd" pointer, increase the
						// character count
						if (wJE < minEnd) {
							minEnd = wJE;
							minEndCharacterDelta = calcCharacterCount(wJT, mode);
						}
					}

					// Advance the current end to the smallest matching word
					if (minEnd < Integer.MAX_VALUE && minEnd > currentEnd) {
						currentEnd = minEnd;
						currentCharacterCount += minEndCharacterDelta;
					}
				}

				// If the current character count is larger or equals the minimum
				// character count, create a new match and continue
				if (currentCharacterCount >= minCharacterCount) {
					final DictionaryMatch match = new DictionaryMatch(node, name);
					match.setBegin(wIB);
					match.setEnd(currentEnd);
					match.setProbability(1.0);
					res.add(match);
				}
			}
		}

		return res;
	}

	public static class DictionaryMatchCompositor implements
			AnnotationCompositor<DictionaryMatch> {

		private final NodeName name;
		private final NodeHandle node;
		private final double pThresh;
		private final int maxCharCount;

		public DictionaryMatchCompositor(NodeName name, NodeHandle node,
				double pThresh, int maxCharCount) {
			this.name = name;
			this.node = node;
			this.pThresh = pThresh;
			this.maxCharCount = maxCharCount;
		}

		@Override
		public DictionaryMatch compose(AnnotationCandidate core,
				ArrayList<ProbableHypothesis> slotCandidates) {
			// Initialize the match.
			final DictionaryMatch currentMatch
					= new DictionaryMatch(node, name);
			final WordReference coreWord = (WordReference) core;
			coreWord.expandDictionaryMatch(currentMatch, coreWord.getMatch().
					getRelativeMatchProbability(maxCharCount));

			// Add all slots
			for (final ProbableHypothesis hypo : slotCandidates) {
				if (hypo != null && hypo.getCandidate() != null) {
					final WordReference wordRef = (WordReference) hypo.
							getCandidate();
					final double p = hypo.getProbability();
					if (wordRef != null && p > 0.0) {
						wordRef.expandDictionaryMatch(currentMatch, p);
					}
				}
			}

			// Abort if the total probability is below the threshold
			if (currentMatch.getProbability() < pThresh) {
				return null;
			}
			return currentMatch;
		}
	}
}
