/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.ontology.driver;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang3.builder.HashCodeBuilder;

/**
 * The NodeName class represents a single "name" that is attached to a ontology
 * node. Each name consists of the actual name and an optional type. This type
 * may contain additional information about the name (e.g. whether this is a
 * colloquial name, a brand name, ...)
 *
 * @author Andreas Stoeckel - astoecke(at)techfak.uni-bielefeld.de
 */
public final class NodeName {

	/**
	 * Contains the actual name of the node.
	 */
	private final String name;

	/**
	 * Specifies the type of the name (e.g. whether this is a colloquial name, a
	 * brand name, ...). May be null.
	 */
	private final String type;

	/**
	 * Specifies the comparison type which is used for comparing the name when
	 * performing dictionary lookups.
	 */
	private final ComparisonMode comparisonMode;

	/**
	 * Specifies whether this node is primary or not.
	 */
	private final boolean primary;

	/**
	 * Creates a new NodeName instance.
	 *
	 * @param name is the actual name of the node.
	 * @param type specifies the type of the name (e.g. whether this is a
	 * colloquial name, a brand name, ...). May be null.
	 * @param comparisonMode defines the comparison mode used when querying the
	 * NodeName from the database
	 * @param primary spcifies whether whis node is primary or not.
	 */
	public NodeName(String name, String type,
			ComparisonMode comparisonMode, boolean primary) {
		this.name = name;
		this.type = type;
		this.comparisonMode = comparisonMode;
		this.primary = primary;
	}

	/**
	 * Creates a new NodeName instance.
	 *
	 * @param name is the actual name of the node.
	 * @param type specifies the type of the name (e.g. whether this is a
	 * colloquial name, a brand name, ...). May be null.
	 * @param comparisonMode defines the comparison mode used when querying the
	 * NodeName from the database
	 */
	public NodeName(String name, String type, ComparisonMode comparisonMode) {
		this(name, type, comparisonMode, false);
	}

	/**
	 * Creates a new NodeName instance. Sets the comparison mode to "fuzzy".
	 *
	 * @param name is the actual name of the node.
	 * @param type specifies the type of the name (e.g. whether this is a
	 * colloquial name, a brand name, ...). May be null.
	 * @param primary spcifies whether whis node is primary or not.
	 */
	public NodeName(String name, String type, boolean primary) {
		this(name, type, ComparisonMode.FUZZY, primary);
	}
	
	/**
	 * Creates a new NodeName instance. Sets the name type to null and the
	 * comparison mode to "fuzzy".
	 *
	 * @param name is the actual name of the node.
	 * @param comparisonMode defines the comparison mode used when querying the
	 * NodeName from the database
	 * @param primary spcifies whether whis node is primary or not.
	 */
	public NodeName(String name, ComparisonMode comparisonMode, boolean primary) {
		this(name, null, comparisonMode, primary);
	}

	/**
	 * Creates a new NodeName instance. Sets the name type to null and the
	 * comparison mode to "fuzzy".
	 *
	 * @param name is the actual name of the node.
	 * @param primary spcifies whether whis node is primary or not.
	 */
	public NodeName(String name, boolean primary) {
		this(name, (String) null, primary);
	}

	/**
	 * Creates a new NodeName instance. Sets the comparison mode to "fuzzy".
	 *
	 * @param name is the actual name of the node.
	 * @param type specifies the type of the name (e.g. whether this is a
	 * colloquial name, a brand name, ...). May be null.
	 */
	public NodeName(String name, String type) {
		this(name, type, ComparisonMode.FUZZY, false);
	}

	/**
	 * Creates a new NodeName instance. Sets the name type to null and the
	 * comparison mode to "fuzzy".
	 *
	 * @param name is the actual name of the node.
	 * @param comparisonMode defines the comparison mode used when querying the
	 * NodeName from the database
	 */
	public NodeName(String name, ComparisonMode comparisonMode) {
		this(name, null, comparisonMode, false);
	}

	/**
	 * Creates a new NodeName instance. Sets the name type to null and the
	 * comparison mode to "fuzzy".
	 *
	 * @param name is the actual name of the node.
	 */
	public NodeName(String name) {
		this(name, (String) null, false);
	}

	/**
	 * Returns the actual name of the node.
	 *
	 * @return the actual name of the node.
	 */
	public String getName() {
		return name;
	}

	/**
	 * Returns the type of the name (e.g. whether this is a colloquial name, a
	 * brand name, ...). May be null.
	 *
	 * @return the type of the name or null.
	 */
	public String getType() {
		return type;
	}

	/**
	 * Returns the comparison mode of the name.
	 *
	 * @return the comparison mode of the name.
	 */
	public ComparisonMode getComparisonMode() {
		return comparisonMode;
	}

	/**
	 * Returns whether this node is primary or not.
	 *
	 * @return whether this node is primary or not.
	 */
	public boolean isPrimary() {
		return primary;
	}

	/**
	 * Returns the name in a string representation of the form "name [type]" if
	 * the type is non-null or returns just the name.
	 *
	 * @return the NodeName in a string form.
	 */
	@Override
	public String toString() {
		return name + ((type != null) ? " [" + type + "]" : "");
	}

	/**
	 * Checks whether both name and type equal.
	 *
	 * @param obj is the object the NodeName instance is compared to.
	 * @return true if both name and type are equal, false otherwise.
	 */
	@Override
	public boolean equals(Object obj) {
		if ((obj == null) || (obj.getClass() != getClass())) {
			return false;
		}

		NodeName n = (NodeName) obj;
		return StringUtils.equals(name, n.name)
				&& StringUtils.equals(type, n.type)
				&& comparisonMode == n.comparisonMode;
	}

	/**
	 * Computes an hash code from both name and type.
	 *
	 * @return a hash code which represents name and type of the NodeName
	 * instance.
	 */
	@Override
	public int hashCode() {
		return (new HashCodeBuilder())
				.append(name)
				.append(type)
				.append(comparisonMode).toHashCode();
	}

}
