/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.ontology.driver;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * The DatabaseDriver class describes an interface that must be implemented by
 * the actual database backend. There is a unit test that tests the
 * implementations of this class for conformance with the interface description.
 *
 * @author Andreas Stoeckel -- astoecke(at)techfak.uni-bielefeld.de
 */
public abstract class OntologyDatabaseDriver {

	/* Utility functions */
	/**
	 * Converts a list of node descriptors to a list of node handles.
	 *
	 * @param nodes is a list of node descriptors.
	 * @return a list of node handles.
	 */
	public static Collection<NodeHandle> nodeHandlesFromDescriptors(
			Collection<NodeDescriptor> nodes) {
		Collection<NodeHandle> handles = new ArrayList(nodes.size());
		for (NodeDescriptor node : nodes) {
			handles.add(node.getHandle());
		}
		return handles;
	}

	/* Ontology management */

	/**
	 * Returns a list containing all ontologies in the database and an empty
	 * list if no ontologies are present. This function must always succeed.
	 *
	 * @return a list containing all available ontologies.
	 */
	public abstract Collection<OntologyHandle> getOntologies();

	/**
	 * Returns a the handle for the ontology with the given name or creates the
	 * ontology if it does not exist. This function must work for all names in
	 * the regexp "[A-Za-Z_][A-Za-z0-9_]*", if another name is provided the
	 * behaviour may be undefined.
	 *
	 * @param name is the name of the ontology that should be returned/created.
	 * @param create if true, the ontology is created if it does not exist yet,
	 * if false, null is returned if the ontology does not exist yet.
	 * @return an ontology handle for an ontology with the given name or null if
	 * the ontology does not exist and create is false.
	 */
	public abstract OntologyHandle getOntology(String name, boolean create);

	/**
	 * Returns the handle for the ontology with the given name or creates the
	 * ontology if it does not exist. This function must work for all names in
	 * the regexp "[A-Za-Z_][A-Za-z0-9_]*", if another name is provided the
	 * behaviour may be undefined.
	 *
	 * @param name is the name of the ontology that should be returned/created.
	 * @return an ontology handle for an ontology with the given name. This
	 * function must always return a valid ontology handle given the name is
	 * valid (fulfills the regexp given above).
	 */
	public OntologyHandle getOntology(String name) {
		return getOntology(name, true);
	}

	/**
	 * Returns the last update date for the given ontology handle.
	 * 
	 * @param ontology is the handle of the ontology for which the modification
	 * date should be returned.
	 * @return the date of the last modification.
	 */
	public abstract Date getOntologyModificationDate(OntologyHandle ontology);

	/**
	 * Returns the last update date for the given ontology handle.
	 * 
	 * @param ontology is the handle of the ontology for which the modification
	 * date should be updated.
	 * @param date is the date the modification date should be set to.
	 */
	public abstract void setOntologyModificationDate(OntologyHandle ontology,
			Date date);

	/**
	 * Deletes the ontology with the given ontology handle. Does nothing if the
	 * ontology does not exist.
	 *
	 * @param ontology is the handle of the ontology that should be deleted.
	 */
	public abstract void deleteOntology(OntologyHandle ontology);

	/**
	 * Returns an empty instance of the ontology with the given name. This
	 * function should be used by importers in order to ensure that the ontology
	 * is actually empty.
	 *
	 * @param name is the name of the ontology that should be returned.
	 * @return is an ontology handle that will be returned and that is ensured
	 * to have no entries.
	 */
	public OntologyHandle getEmptyOntology(String name) {
		OntologyHandle ontology = getOntology(name, false);
		if (ontology != null) {
			deleteOntology(ontology);
		}
		return getOntology(name);
	}

	/* Node management */
	/**
	 * Returns a list containing all root nodes in the given ontology.
	 *
	 * @param ontology is the ontology for which the root nodes should be
	 * returned.
	 * @return a list containing the nodes that are root nodes in the given
	 * ontology, an empty list if no such nodes exist or null if the given
	 * ontology does not exist.
	 */
	public abstract Collection<NodeDescriptor> getRootNodes(
			OntologyHandle ontology);

	/**
	 * Returns the node descriptors for the given node handles.
	 *
	 * @param handles is a list of node handles.
	 * @return a list of node descriptors. If the given collection is a list the
	 * order must be obtained. Non-existing entires have to be inserted as
	 * "null". It must hold result.size() == handles.size().
	 */
	public abstract List<NodeDescriptor> getNodes(Collection<NodeHandle> handles);

	/**
	 * Returns the node descriptor for a single node.
	 *
	 * @param node is the node that should be looked up in the database.
	 * @return a node descriptor or null if the given node does not exist.
	 */
	public NodeDescriptor getNode(NodeHandle node) {
		return getNodes(Arrays.asList(new NodeHandle[]{node})).get(0);
	}

	/**
	 * Checks whether the given nodes exist.
	 *
	 * @param handles is a list of node handles.
	 * @return a list containing "true" if the corresponding node exists or
	 * "false" if it doesn't. If the given collection is a list the order must
	 * be obtained. It must hold result.size() == handles.size().
	 */
	public abstract List<Boolean> queryNodesExist(Collection<NodeHandle> handles);

	/**
	 * Checks whether all given nodes exist.
	 *
	 * @param handles is a list of node handles.
	 * @return true if all given nodes exist (AND), false if at least one of the
	 * given nodes does not exist.
	 */
	public boolean queryAllNodesExist(Collection<NodeHandle> handles) {
		boolean allExist = true;
		for (Boolean b : queryNodesExist(handles)) {
			allExist = allExist && b;
		}
		return allExist;
	}

	/**
	 * Checks whether the given node handle is valid.
	 *
	 * @param handle is a node handle.
	 * @return true if the corresponding node really exists, false otherwise.
	 */
	public boolean queryNodeExists(NodeHandle handle) {
		return queryNodesExist(Arrays.asList(new NodeHandle[]{handle})).get(0);
	}

	/**
	 * Stores the given nodes in the database.
	 *
	 * @param descriptors is a list of node descriptors that should be stored.
	 * Invalid node descriptors (e.g. refering to a non existing ontology) must
	 * be ignored.
	 *
	 * TODO: Check this in the unit test/implementation, as this is not done at
	 * the moment (would probably be to costly).
	 */
	public abstract void storeNodes(Collection<NodeDescriptor> descriptors);

	/**
	 * Stores a single node in the database.
	 *
	 * @param node is the node that should be stored.
	 */
	public void storeNode(NodeDescriptor node) {
		storeNodes(Arrays.asList(new NodeDescriptor[]{node}));
	}

	/**
	 * Deletes the given nodes from the database. Does nothing if one of the
	 * given node handles is invalid (node/ontology does not exist).
	 *
	 * @param nodes is a list containing the nodes that should be deleted.
	 */
	public abstract void deleteNodes(Collection<NodeHandle> nodes);

	/**
	 * Deletes a single node from the database. Does nothing if the node does
	 * not exist.
	 *
	 * @param node is the node that should be deleted.
	 */
	public void deleteNode(NodeHandle node) {
		deleteNodes(Arrays.asList(new NodeHandle[]{node}));
	}

	/**
	 * Deletes the given nodes from the database. Does nothing if one of the
	 * given node handles is invalid (node/ontology does not exist).
	 *
	 * @param nodes is a list containing the nodes that should be deleted.
	 */
	public void deleteNodeDescriptors(Collection<NodeDescriptor> nodes) {
		Collection<NodeHandle> handles = new ArrayList(nodes.size());
		for (NodeDescriptor node : nodes) {
			handles.add(node.getHandle());
		}
		deleteNodes(handles);
	}

	/**
	 * Returns the associations for a given set of nodes.
	 *
	 * @param nodes is a list of nodes.
	 * @param associationTypes is a bit field containing the association types
	 * that should be returned.
	 * @param dir is the association direction that should be returned.
	 * @return a list of collections of associations. If a node has no
	 * associations, the corresponding list should be empty. If a node is
	 * invalid, the list entry should be null. If the given collection is a list
	 * the order must be obtained. It must hold result.size() == nodes.size().
	 *
	 * TODO: Test this
	 */
	public abstract List<Collection<Association>> getAssociations(
			Collection<NodeHandle> nodes, int associationTypes,
			AssociationDirection dir);

	/**
	 * Returns the associations for a single node.
	 *
	 * @param node is the node for which the associations should be returned.
	 * @param associationTypes is a bit field containing the association types
	 * that should be returned.
	 * @param dir is the association direction that should be returned.
	 * @return a collection of associations. If a node has no associations, the
	 * collection should be empty. If a node is invalid, the list entry should
	 * be null.
	 *
	 * TODO: Test this
	 */
	public Collection<Association> getAssociations(NodeHandle node,
			int associationTypes, AssociationDirection dir) {
		return getAssociations(Arrays.asList(new NodeHandle[]{node}),
				associationTypes, dir).get(0);
	}

	/**
	 * Stores the given associations.
	 *
	 * @param associations is a list of associations that should be stored. The
	 * database does not have to check whether the associations are actually
	 * valid.
	 *
	 * TODO: Change this?
	 */
	public abstract void storeAssociations(Collection<Association> associations);

	/**
	 * Stores the given associations. The two nodes in the association have to
	 * belong to the same ontology. If a node does not exist yet, it is created.
	 *
	 * @param association is an association that should be stored. The database
	 * does not have to check whether the associations are valid.
	 */
	public void storeAssociation(Association association) {
		storeAssociations(Arrays.asList(new Association[]{association}));
	}

	/**
	 * Deletes the given associations. Does nothing if an association does not
	 * exist.
	 *
	 * @param associations is a list of associations that should be deleted.
	 */
	public abstract void deleteAssociations(Collection<Association> associations);

	/**
	 * Deletes the given association. Does nothing if the association does not
	 * exist.
	 *
	 * @param association the association that should be deleted.
	 */
	public void deleteAssociation(Association association) {
		deleteAssociations(Arrays.asList(new Association[]{association}));
	}

	/* Dictionary managment */
	/**
	 * Queries the dictionary for the given keys.
	 *
	 * @param keys contains a set of keys that should be queried.
	 * @return a map of the queried keys to collections of associated nodes.
	 * Keys with no result may be ommited from the map.
	 */
	public abstract Map<String, Collection<NodeDescriptor>> queryDictionary(
			Set<String> keys);

	/**
	 * Queries the dictionary for a given key.
	 *
	 * @param key contains a set of keys that should be queried.
	 * @return a collection of nodes that correspond to the key, null or an
	 * empty collection if the key has no associated nodes.
	 */
	public Collection<NodeDescriptor> queryDictionary(String key) {
		return (Collection<NodeDescriptor>) queryDictionary(
				new HashSet(Arrays.asList(new String[]{key}))).get(key);
	}

	/**
	 * Stores the given node to key associations in the dictionary.
	 *
	 * @param keys is a map from node handles to keys that should be stored in
	 * the dictionary.
	 */
	public abstract void storeDictionaryEntries(
			Map<NodeHandle, Set<String>> keys);

	/**
	 * Stores the given node to key association in the dictionary.
	 *
	 * @param handle is the handle for which the keys should be stored in the
	 * db.
	 * @param keys is list of keys that should be associated to the given
	 * handles.
	 */
	public void storeDictionaryEntry(NodeHandle handle, Set<String> keys) {
		Map<NodeHandle, Set<String>> tmp = new HashMap();
		tmp.put(handle, keys);
		storeDictionaryEntries(tmp);
	}

	/**
	 * Deletes the dictionary entries for the given nodes from the db.
	 *
	 * @param nodes is a list of nodes for which the dictionary entries should
	 * be deleted.
	 */
	public abstract void deleteDictionaryEntries(Collection<NodeHandle> nodes);

	/**
	 * Deletes the dictionary entries for a single node.
	 *
	 * @param node is the node for which the dictionary entries should be
	 * deleted.
	 */
	public void deleteDictionaryEntry(NodeHandle node) {
		deleteDictionaryEntries(Arrays.asList(new NodeHandle[]{node}));
	}

	/**
	 * Gives the driver the opportunity to shut down properly.
	 */
	public void close() {
		// Do nothing here.
	}

	/**
	 * Gives the driver the opportunity to return a database version number.
	 * The returned string should be of the form "YYYY-MM-DD.Rev" where "Rev"
	 * is a two digit revision number for the given date.
	 */
	public String getSchemaVersion() {
		return "0000-00-00.00";
	}

	/**
	 * Wipes the ontology database clean.
	 */
	public abstract void clear();
}
