/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.ontology.importer;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;

import de.unibi.techfak.scie.ontology.OntologyDatabase;
import de.unibi.techfak.scie.ontology.driver.NodeName;
import de.unibi.techfak.scie.ontology.driver.OntologyHandle;
import java.util.Date;

public abstract class AbstractImporter {

	/**
	 * Reference to the ontology database into which the data will be imported.
	 */
	protected final OntologyDatabase db;

	/**
	 * Name of the ncbi ontology.
	 */
	protected final String ontologyName;

	/**
	 * Instantiates the importer for the given database.
	 *
	 * @param db is a reference to the database in which the ontology should be
	 * stored.
	 * @param ontologyName is the name of the ontology in which the data will be
	 * stored.
	 */
	public AbstractImporter(OntologyDatabase db, String ontologyName) {
		this.db = db;
		this.ontologyName = ontologyName;
	}

	/**
	 * Flushes all data from xml into db.
	 *
	 * @param parser , the XMLStreamReader for the Drugbank xml
	 * @param onto, the OntologyHandle
	 * @throws XMLStreamException
	 */
	protected abstract void parseXML(OntologyHandle onto, XMLStreamReader parser) throws XMLStreamException;

	public void importOntology(File xmlFile, boolean override) {
		// Fetch the ontology -- check whether the given file is younger than
		// the stored ontology information.
		if (db.queryOntologyExists(ontologyName) && !override) {
			Date ontoDate = db.getOntologyModificationDate(db.getOntology(ontologyName));
			Date fileDate = new Date(xmlFile.lastModified());
			if (ontoDate.after(fileDate)) {
				System.out.println("The ontology " + ontologyName + " exists "
						+ "and is up to date (File date: " + fileDate
						+ " OntologyDate:" + ontoDate + "). Add \"override\" as "
						+ "an parameter if you want to override the existing "
						+ "ontology.");
				return;
			}
		}
		OntologyHandle onto = db.getEmptyOntology(ontologyName);

		// Create the input file parser
		try {
			XMLInputFactory factory = XMLInputFactory.newInstance();
			factory.setProperty(XMLInputFactory.SUPPORT_DTD, false); // Workaround for Bug #747 that appeared out of thin air (!?)
			XMLStreamReader parser = factory.createXMLStreamReader(new FileInputStream(xmlFile));
			parseXML(onto, parser);
		} catch (FileNotFoundException ex) {
			System.err.println("The given file \"" + xmlFile.getPath() + "\" was not found. Skipping.");
		} catch (XMLStreamException ex) {
			throw new RuntimeException("Error while reading File.", ex);
		}
	}

	protected static final Pattern[] CLEAN_REDUCE = new Pattern[]{
		Pattern.compile("(.*)(sp|cf|aff)\\..*$"),
		Pattern.compile("(.*) in .*, [1-9][0-9][0-9][0-9]$"),
		Pattern.compile("(.*) [^\\s]+ & [^\\s]+, [1-9][0-9][0-9][0-9]$"),
		Pattern.compile("(.*) [^\\s]+, [1-9][0-9][0-9][0-9]$")
	};

	protected static final Pattern REDUCE_SPACES = Pattern.compile("(^\\s+|\\s+$|\\s\\s+)");

	protected static final Pattern CLEAN_PARANTHESIS = Pattern.compile("(\\(.*\\)| '.*'| \".*\")");

}
