/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.util;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.NavigableMap;
import java.util.TreeMap;
import org.apache.uima.fit.util.JCasUtil;
import org.apache.uima.jcas.JCas;
import org.apache.uima.jcas.tcas.Annotation;

/**
 * This is a custom wrapper implementation to reduce the necessary calls to UIMA JCasUtil, which
 * seems to be terribly slow.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class CachedJCasUtil {

	private final JCas jcas;

	private CachedJCasUtil(JCas jcas) {
		this.jcas = jcas;
	}

	public JCas getJcas() {
		return jcas;
	}
	/**
	 * This is our current Singleton. Please note that this Singleton is replaced each time a new
	 * JCas instance is used. This should be changed if a parallel processing is implemented or
	 * something like that. This implicitly assumes that JCas instances are processed sequentially.
	 *
	 * We choose this implementation for now because it ensures that our working memory does not
	 * flood with cached instances for each and every JCas instance.
	 */
	private static CachedJCasUtil currentInstance;

	/**
	 * Returns the JCasUtil instance for the given JCas instance.
	 *
	 * @param jcas
	 * @return
	 */
	public static CachedJCasUtil getInstance(final JCas jcas) {
		if (currentInstance == null || jcas != currentInstance.getJcas()) {
			currentInstance = new CachedJCasUtil(jcas);
		}
		return currentInstance;
	}
	/**
	 * We store a Tree-based cache for each class that is queried.
	 *
	 * Each cache stores the Annotations for a given gives in order of their starting position.
	 */
	private HashMap<Class<? extends Annotation>, AnnotationCache> caches =
			new HashMap<>();

	/**
	 * Manually adds the given instance to the respective cache.
	 *
	 * Please note that this constructs the cache if it is not there yet. During construction all
	 * Annotions of the given class will be added to the cache. If the given Annotation is part of
	 * the UIMA indices already, it will thus be added twice.
	 *
	 * In short: Please do only use this method if you know, what you are doing. Otherwise use
	 * "update" after you have added new Annotations.
	 *
	 * @param <X>
	 * @param annotationInstance
	 */
	public <X extends Annotation> void add(X annotationInstance) {
		add((Class<X>) annotationInstance.getClass(), annotationInstance);
	}

	/**
	 * Manually adds the given instance to the respective cache.
	 *
	 * Please note that this constructs the cache if it is not there yet. During construction all
	 * Annotions of the given class will be added to the cache. If the given Annotation is part of
	 * the UIMA indices already, it will thus be added twice.
	 *
	 * In short: Please do only use this method if you know, what you are doing. Otherwise use
	 * "update" after you have added new Annotations.
	 *
	 * @param <X>
	 * @param annotationInstance
	 */
	public <X extends Annotation> void add(Class<X> annotationClass, X annotationInstance) {
		//get the cache
		final AnnotationCache<X> cache = getCache(annotationClass);
		cache.add(annotationInstance);
	}

	public <X extends Annotation> boolean remove(Class<X> annotationClass, X annotationInstance) {
		//get the cache
		final AnnotationCache<X> cache = getCache(annotationClass);
		return cache.remove(annotationInstance);
	}

	/**
	 * Updates the cache for the given class and inserts all new instances of the given class.
	 *
	 * @param <X>
	 * @param annotationClass
	 */
	public <X extends Annotation> void update(Class<X> annotationClass) {
		final Collection<X> annoInstances = JCasUtil.select(jcas, annotationClass);
		final AnnotationCache<X> cache = new AnnotationCache<>();
		for (final X anno : annoInstances) {
			cache.add(anno);
		}
		caches.put(annotationClass, cache);
	}

	/**
	 * Returns all annotations of the given class.
	 *
	 * The returned list has a well defined order: The output Annotations are first ordered
	 * according to their begin, than according to their end index.
	 */
	public <X extends Annotation> List<X> select(Class<X> annotationClass) {
		final AnnotationCache<X> cache = getCache(annotationClass);
		return cache.getAll();
	}

	/**
	 * Returns all annotations of the given class that are contained in the text region specified by
	 * the given begin and end (inclusive).
	 *
	 * The returned list has a well defined order: The output Annotations are first ordered
	 * according to their begin, than according to their end index.
	 */
	public <X extends Annotation> List<X> selectCovered(Class<X> annotationClass, int begin, int end) {
		final AnnotationCache<X> cache = getCache(annotationClass);
		return cache.getCovered(begin, end);
	}

	/**
	 * Returns all annotations of the given class that contain at least the text region specified by
	 * the given begin and end (inclusive).
	 *
	 * The returned list has a well defined order: The output Annotations are first ordered
	 * according to their begin, than according to their end index.
	 */
	public <X extends Annotation> List<X> selectCovering(Class<X> annotationClass, int begin,
			int end) {
		final AnnotationCache<X> cache = getCache(annotationClass);
		return cache.getCovering(begin, end);
	}
	
	/**
	 * Returns all annotations of the given class that are overlapping with the text region specified by
	 * the given begin and end (inclusive).
	 *
	 * The returned list has a well defined order: The output Annotations are first ordered
	 * according to their begin, than according to their end index.
	 */
	public <X extends Annotation> List<X> selectOverlapping(Class<X> annotationClass, int begin,
			int end) {
		final AnnotationCache<X> cache = getCache(annotationClass);
		return cache.getOverlapping(begin, end);
	}

	/**
	 * This constructs a cache for the given class if necessary or returns it if it is already
	 * cached.
	 */
	private <X extends Annotation> AnnotationCache<X> getCache(Class<X> annotationClass) {
		AnnotationCache<X> cache = caches.get(annotationClass);
		if (cache == null) {
			final Collection<X> annoInstances = JCasUtil.select(jcas, annotationClass);
			cache = new AnnotationCache<>();
			for (final X anno : annoInstances) {
				cache.add(anno);
			}
			caches.put(annotationClass, cache);
		}
		return cache;
	}

	private static class AnnotationCache<X extends Annotation> {

		/**
		 * We store the annotations themselves in a TreeMap indiced by their start position. For the
		 * case of overlapping annotations we have an additional hierarchy level of a TreeMap that
		 * stores the Annotations with the same start positions indiced according to their end
		 * position. Unfortunately, even that is not enough, because several annotations of the same
		 * class can exist for the same begin and end index. Thus we need a LinkedList at that last
		 * level of hierarchy.
		 */
		private final TreeMap<Integer, TreeMap<Integer, LinkedList<X>>> actualCashe =
				new TreeMap<>();
		/**
		 * For the special case of overlapping annotations however we also need to now how many
		 * annotations we need to check in linear time to ensure that given criteria are met.
		 */
		private int maxLength = 0;

		public AnnotationCache() {
		}

		public void add(X anno) {
			final int length = anno.getEnd() - anno.getBegin();
			if (length > maxLength) {
				maxLength = length;
			}
			TreeMap<Integer, LinkedList<X>> overlappingAnnos = actualCashe.get(anno.getBegin());
			if (overlappingAnnos == null) {
				overlappingAnnos = new TreeMap<>();
				actualCashe.put(anno.getBegin(), overlappingAnnos);
			}
			LinkedList<X> samePosAnnos = overlappingAnnos.get(anno.getEnd());
			if (samePosAnnos == null) {
				samePosAnnos = new LinkedList<>();
				overlappingAnnos.put(anno.getEnd(), samePosAnnos);
			}

			samePosAnnos.add(anno);
		}

		public boolean remove(X anno) {
			TreeMap<Integer, LinkedList<X>> overlappingAnnos = actualCashe.get(anno.getBegin());
			if (overlappingAnnos == null) {
				return false;
			}
			LinkedList<X> samePosAnnos = overlappingAnnos.get(anno.getEnd());
			if (samePosAnnos == null) {
				return false;
			}
			return samePosAnnos.remove(anno);
		}

		/**
		 * Returns all Annotations. The returned list has a well defined order: The output
		 * Annotations are first ordered according to their begin, than according to their end
		 * index.
		 *
		 * @return
		 */
		public List<X> getAll() {
			final ArrayList<X> returnList = new ArrayList<>(actualCashe.size());
			for (TreeMap<Integer, LinkedList<X>> overlappingAnnos : actualCashe.values()) {
				for (LinkedList<X> samePosAnnos : overlappingAnnos.values()) {
					returnList.addAll(samePosAnnos);
				}
			}
			return returnList;
		}

		/**
		 * Returns all annotations stored in this cache that start before (or at) the start index
		 * and end after (or at) the end index.
		 *
		 * The returned list has a well defined order: The output Annotations are first ordered
		 * according to their begin, than according to their end index.
		 *
		 * @param start
		 * @param end
		 * @return
		 */
		public List<X> getCovering(int start, int end) {
			final ArrayList<X> returnList = new ArrayList<>();
			final int lowerKey = end - maxLength;
			if (lowerKey > start) {
				return returnList;
			}
			/*
			 * We only care about the SubMap of Annotations that begin from
			 * 
			 * end-maxLength
			 * 
			 * to
			 * 
			 * start
			 * 
			 */
			final NavigableMap<Integer, TreeMap<Integer, LinkedList<X>>> subMap =
					actualCashe.subMap(lowerKey, true, start, true);
			//for that subMap we check if the given criteria are met.
			for (final TreeMap<Integer, LinkedList<X>> overlappingAnnos : subMap.values()) {
				//get only the annotations that end at or after the specified end.
				final NavigableMap<Integer, LinkedList<X>> samePosAnnoMap = overlappingAnnos.
						tailMap(end, true);
				//and add them all.
				for (LinkedList<X> samePosAnnos : samePosAnnoMap.values()) {
					returnList.addAll(samePosAnnos);
				}
			}
			return returnList;
		}

		/**
		 * Returns all annotations stored in this cache that start after (or at) the start index and
		 * end before (or at) the end index.
		 *
		 * The returned list has a well defined order: The output Annotations are first ordered
		 * according to their begin, than according to their end index.
		 *
		 * @param start
		 * @param end
		 * @return
		 */
		public List<X> getCovered(int start, int end) {
			/*
			 * We only care about the SubMap of Annotations that begin from
			 * 
			 * begin
			 * 
			 * to
			 * 
			 * end
			 */
			final NavigableMap<Integer, TreeMap<Integer, LinkedList<X>>> subMap =
					actualCashe.subMap(start, true, end, true);
			final ArrayList<X> returnList = new ArrayList<>();
			//for that subMap we check if the given criteria are met.
			for (final TreeMap<Integer, LinkedList<X>> overlappingAnnos : subMap.values()) {
				//get only the annotations that end before or at the specified end.
				final NavigableMap<Integer, LinkedList<X>> samePosAnnoMap = overlappingAnnos.
						headMap(end, true);
				//and add them all.
				for (LinkedList<X> samePosAnnos : samePosAnnoMap.values()) {
					returnList.addAll(samePosAnnos);
				}
			}
			return returnList;
		}
		
		/**
		 * Returns all annotations stored in this cache that start before (or at) the end index and
		 * end after (or at) the start index.
		 *
		 * The returned list has a well defined order: The output Annotations are first ordered
		 * according to their begin, than according to their end index.
		 *
		 * @param start
		 * @param end
		 * @return
		 */
		public List<X> getOverlapping(int start, int end) {
			final ArrayList<X> returnList = new ArrayList<>();
			final int lowerKey = end - maxLength;
			if (lowerKey > start) {
				return returnList;
			}
			/*
			 * We only care about the SubMap of Annotations that begin from
			 * 
			 * end-maxLength
			 * 
			 * to
			 * 
			 * end
			 * 
			 */
			final NavigableMap<Integer, TreeMap<Integer, LinkedList<X>>> subMap =
					actualCashe.subMap(lowerKey, true, end, true);
			//for that subMap we check if the given criteria are met.
			for (final TreeMap<Integer, LinkedList<X>> overlappingAnnos : subMap.values()) {
				//get only the annotations that end at or after the specified end.
				final NavigableMap<Integer, LinkedList<X>> samePosAnnoMap = overlappingAnnos.
						tailMap(start, true);
				//and add them all.
				for (LinkedList<X> samePosAnnos : samePosAnnoMap.values()) {
					returnList.addAll(samePosAnnos);
				}
			}
			return returnList;
		}

		public List<X> getFollowing(int index, int count) {
			//get the sub map following the given index.
			final NavigableMap<Integer, TreeMap<Integer, LinkedList<X>>> tailMap =
					actualCashe.tailMap(index, false);
			final ArrayList<X> returnList = new ArrayList<>();
			for (final TreeMap<Integer, LinkedList<X>> overlappingAnnos : tailMap.values()) {
				for (final LinkedList<X> samePosAnnos : overlappingAnnos.values()) {
					for (final X anno : samePosAnnos) {
						if (returnList.size() == count) {
							return returnList;
						}
						returnList.add(anno);
					}
				}
			}
			return returnList;

		}
	}
}
