/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.annotators.animal;

import de.unibi.techfak.scie.annotators.animal.WeightAnnotator;
import de.unibi.techfak.scie.annotators.quantities.NumberAnnotator;
import de.unibi.techfak.scie.annotators.quantities.RangeAnnotator;
import de.unibi.techfak.scie.annotators.quantities.Unit;
import de.unibi.techfak.scie.annotators.quantities.WordNumberAnnotator;
import de.unibi.techfak.scie.annotators.structure.StanfordTokenizer;
import de.unibi.techfak.isy.lppba.descriptors.Weight;
import java.util.Collection;
import java.util.Iterator;
import static junit.framework.Assert.assertEquals;
import static junit.framework.Assert.assertFalse;
import static junit.framework.Assert.assertTrue;
import junit.framework.TestCase;
import org.apache.uima.UIMAException;
import org.apache.uima.fit.factory.JCasFactory;
import org.apache.uima.fit.util.JCasUtil;
import org.apache.uima.jcas.JCas;

/**
 *
 * @author Benjamin Paassen - bpaassen(at)techfak.uni-bielefeld.de
 */
public class WeightAnnotatorTest extends TestCase {

	public WeightAnnotatorTest(String testName) {
		super(testName);
	}

	@Override
	protected void setUp() throws Exception {
		super.setUp();
	}

	@Override
	protected void tearDown() throws Exception {
		super.tearDown();
	}

	/**
	 * Test of convertToSIValue method, of class WeightAnnotator.
	 */
	public void testConvertToSIValue() {
		assertEquals(1000, WeightAnnotator.convertToSIValue(Unit.TONNES, 1),0);
		assertEquals(1, WeightAnnotator.convertToSIValue(Unit.KILOGRAMMES, 1),0);
		assertEquals(1, WeightAnnotator.convertToSIValue(Unit.GRAMMES, 1000),0);
		assertEquals(1, WeightAnnotator.convertToSIValue(Unit.MILLIGRAMMES, 1000000),0);
	}

	/**
	 * Test of handleQuantity method, of class WeightAnnotator.
	 */
	public void testHandleQuantity() throws UIMAException {

		//create example
		final JCas jcas = JCasFactory.createJCas(
				"src/main/resources/de/unibi/techfak/isy/lppba/descriptors/LPPBATypeSystem");
		jcas.setDocumentText("Rats weighing 200-300 g. We dropped 1 t of weight. 200 mg of stuff.");
		//preprosess using StanfordTokenizer, NumberAnnotator, RangeAnnotator and WordNumberAnnotator
		final StanfordTokenizer st = new StanfordTokenizer();
		st.process(jcas);
		final NumberAnnotator numberAnnotator = new NumberAnnotator();
		numberAnnotator.process(jcas);
		final WordNumberAnnotator wordNumberAnnotator = new WordNumberAnnotator();
		wordNumberAnnotator.process(jcas);
		final RangeAnnotator rangeAnnotator = new RangeAnnotator();
		rangeAnnotator.process(jcas);

		//use the WeightAnnotator
		final WeightAnnotator instance = new WeightAnnotator();
		instance.process(jcas);

		//check results
		final Collection<Weight> weightInstances = JCasUtil.select(jcas, Weight.class);

		final Iterator<Weight> iterator = weightInstances.iterator();
		assertTrue(iterator.hasNext());
		Weight weight = iterator.next();
		assertEquals("200-300 g", weight.getCoveredText());
		assertEquals("200-300", weight.getValue().getCoveredText());
		assertEquals("Grammes", weight.getUnitName());

		assertTrue(iterator.hasNext());
		weight = iterator.next();
		assertEquals("1 t", weight.getCoveredText());
		assertEquals(1000, weight.getSIValue(),0);
		assertEquals("1", weight.getValue().getCoveredText());
		assertEquals("Tonnes", weight.getUnitName());

		assertTrue(iterator.hasNext());
		weight = iterator.next();
		assertEquals("200 mg", weight.getCoveredText());
		assertEquals(200E-6, weight.getSIValue(),0);
		assertEquals("200", weight.getValue().getCoveredText());
		assertEquals("Milligrammes", weight.getUnitName());

		assertFalse(iterator.hasNext());
	}
	
	/**
	 * Test of handleQuantity method, of class WeightAnnotator.
	 */
	public void testHandleQuantity2() throws UIMAException {

		//create example
		final JCas jcas = JCasFactory.createJCas(
				"src/main/resources/de/unibi/techfak/isy/lppba/descriptors/LPPBATypeSystem");
		jcas.setDocumentText("Forty-two adult female Sprague-Dawley rats (225–250g, Charles River, St-Constant, Quebec, Canada)were used in this study.");
		//preprosess using StanfordTokenizer, NumberAnnotator, RangeAnnotator and WordNumberAnnotator
		final StanfordTokenizer st = new StanfordTokenizer();
		st.process(jcas);
		final NumberAnnotator numberAnnotator = new NumberAnnotator();
		numberAnnotator.process(jcas);
		final WordNumberAnnotator wordNumberAnnotator = new WordNumberAnnotator();
		wordNumberAnnotator.process(jcas);
		final RangeAnnotator rangeAnnotator = new RangeAnnotator();
		rangeAnnotator.process(jcas);

		//use the WeightAnnotator
		final WeightAnnotator instance = new WeightAnnotator();
		instance.process(jcas);

		//check results
		final Collection<Weight> weightInstances = JCasUtil.select(jcas, Weight.class);

		final Iterator<Weight> iterator = weightInstances.iterator();
		assertTrue(iterator.hasNext());
		Weight weight = iterator.next();
		assertEquals("225–250g", weight.getCoveredText());
		assertEquals("225–250", weight.getValue().getCoveredText());
		assertEquals("Grammes", weight.getUnitName());
		assertFalse(iterator.hasNext());
	}
}
