/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.ontology;

import de.unibi.techfak.scie.ontology.Node;
import de.unibi.techfak.scie.ontology.InvalidNameException;
import de.unibi.techfak.scie.ontology.DictionaryMatch;
import de.unibi.techfak.scie.ontology.OntologyDatabase;
import de.unibi.techfak.scie.ontology.driver.Association;
import de.unibi.techfak.scie.ontology.driver.OntologyDatabaseDriver;
import de.unibi.techfak.scie.ontology.driver.OntologyDatabaseDriverTestData;
import de.unibi.techfak.scie.ontology.driver.NodeDescriptor;
import de.unibi.techfak.scie.ontology.driver.NodeHandle;
import de.unibi.techfak.scie.ontology.driver.OntologyHandle;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import static junit.framework.Assert.assertEquals;
import static junit.framework.Assert.assertTrue;
import junit.framework.TestCase;
import org.apache.commons.lang3.ArrayUtils;

/**
 *
 * @author andreas
 */
public abstract class OntologyDatabaseTest extends TestCase {

	public OntologyDatabaseTest(String testName) {
		super(testName);
	}

	abstract protected OntologyDatabaseDriver getTestDriver();

	private static void testDictionaryQuery(OntologyDatabase db, String word,
			int[] ids, boolean strict) {
		Collection<DictionaryMatch> matches = db.query(word);
		int[] tmpIds = ArrayUtils.clone(ids);
		if (strict) {
			assertEquals("match count", ids.length, matches.size());
		}
		for (DictionaryMatch match : matches) {
			System.out.println("\"" + word + "\" matches \"" + match.getName()
					+ "\" (from character " + match.getBegin() + " to "
					+ match.getEnd() + ") with probability "
					+ match.getProbability());
			final int id = match.getNode().getId();
			final int idx = ArrayUtils.indexOf(tmpIds, id);
			if (idx >= 0) {
				tmpIds[idx] = -1;
			}
		}

		// Make sure all required matches were found
		for (int id : tmpIds) {
			assertTrue("Index found", id == -1);
		}
	}

	private static void testDictionaryQuery(OntologyDatabase db, String word,
			int[] ids) {
		testDictionaryQuery(db, word, ids, false);
	}

	public void testOntologies() {
		System.out.println("testOntologies");

		OntologyDatabase db = new OntologyDatabase(getTestDriver());

		// There should be no ontologies
		assertEquals("ontology count", 0, db.getOntologies().size());

		// Create three new ontologies
		OntologyHandle ontology1 = db.getOntology("Onto_1");
		OntologyHandle ontology2 = db.getOntology("Onto_2");
		OntologyHandle ontology3 = db.getOntology("Onto_3");

		// Try to create an ontology with an invalid name -- should throws an
		// exception
		boolean hadException = false;
		try {
			db.getOntology("$:-#");
		} catch (InvalidNameException e) {
			hadException = true;
		}
		assertTrue("invalid name exception", hadException);

		hadException = false;
		try {
			db.getOntology("0bla");
		} catch (InvalidNameException e) {
			hadException = true;
		}
		assertTrue("invalid name exception", hadException);

		// Querying the same ontology should return a handle that is equivalent
		// to the old one
		OntologyHandle ontology3_2 = db.getOntology("Onto_3");
		assertEquals("same handle", ontology3, ontology3_2);

		// There should be three ontologies
		assertEquals("ontology count", 3, db.getOntologies().size());

		// Add nodes to all three ontologies
		db.storeNode(new Node(db, ontology1, 15, "phantom warlord", true));
		db.storeNode(new Node(db, ontology1, 1, "strangling vines"));
		db.storeNode(new Node(db, ontology2, 0, "ruby sword"));
		db.storeNode(new Node(db, ontology2, 1, "enchanted bunny"));
		db.storeNode(new Node(db, ontology3, 0, "the carrot of power", true));
		db.storeNode(new Node(db, ontology3, 1, "two-headed tiger"));

		// Querying the dictionary for "carrot" should work
		testDictionaryQuery(db, "carrot", new int[]{0}, true);

		// Querying the root nodes of ontology1 should return node 15
		List<Node> rootNodes = new ArrayList(db.getRootNodes(ontology1));
		assertEquals("root node count", 1, rootNodes.size());
		assertEquals("root node id", 15, rootNodes.get(0).getId());
		assertEquals("root node name", "phantom warlord", rootNodes.get(0).getPrimaryName());

		// Delete the entire ontology
		db.deleteOntology(ontology3);

		// We should now no longer be able to query the node
		testDictionaryQuery(db, "carrot", new int[]{}, true);

		// The root nodes for this ontology are not available
		assertEquals("root nodes", null, db.getRootNodes(ontology3));

		// The same should happen if we delete ontology1
		db.deleteOntology(ontology1);
		assertEquals("root nodes", null, db.getRootNodes(ontology1));

		// Querying node 15 should now return null
		assertEquals("deleted node", null, db.getNode(new NodeHandle(ontology1, 15)));
	}

	public void testAssociations() {
		System.out.println("testAssociations");

		// Create the test database
		OntologyDatabase db = new OntologyDatabase(getTestDriver());
		OntologyHandle onto = db.getOntology("oiski_poiski");		

		// Store all nodes
		for (NodeDescriptor node : OntologyDatabaseDriverTestData.getTestDescriptors(onto)) {
			db.storeNode(node);
		}

		// Store all associations
		for (Association assoc : OntologyDatabaseDriverTestData.getTestAssociations(onto)) {
			db.getNode(assoc.getSource()).addAssociationTo(assoc.getTarget(), assoc.getType());
		}

		// Query node 7 and get all associated nodes
		Node node7 = db.getNode(new NodeHandle(onto, 7));
		Collection<Node> nodes = node7.getAssociatedNodes();

		// Make sure all expected associations are there	
		assertEquals("Node count", 3, nodes.size());
		int ids[] = new int[]{3, 4, 37};
		for (NodeDescriptor node : nodes) {
			int idx = ArrayUtils.indexOf(ids, node.getId());
			assertEquals("Node ontology", onto, node.getOntology());
			assertTrue("Node id", idx >= 0);
			assertTrue("Node id", ids[idx] != -1);
			ids[idx] = -1;
		}

		// Deleting node 37 should reduce the association count
		db.deleteNode(new NodeHandle(onto, 37));
		nodes = node7.getAssociatedNodes();
		assertEquals("Node count", 2, nodes.size());
		ids = new int[]{3, 4};
		for (NodeDescriptor node : nodes) {
			int idx = ArrayUtils.indexOf(ids, node.getId());
			assertEquals("Node ontology", onto, node.getOntology());
			assertTrue("Node id", idx >= 0);
			assertTrue("Node id", ids[idx] != -1);
			ids[idx] = -1;
		}
	}

	public void testDictionarySimple() {
		System.out.println("testDictionarySimple");

		OntologyDatabase db = new OntologyDatabase(getTestDriver());
		OntologyHandle ontology = db.getOntology("testOntology");

		// Create some nodes
		Node[] nodes = new Node[]{
			new Node(db, ontology, 0, "phantom warlord"),
			new Node(db, ontology, 1, "strangling vines"),
			new Node(db, ontology, 2, "ruby sword"),
			new Node(db, ontology, 3, "enchanted bunny"),
			new Node(db, ontology, 4, "the carrot of power"),
			new Node(db, ontology, 5, "two-headed tiger"),
			new Node(db, ontology, 6, "three-headed lion"),
			new Node(db, ontology, 7, "mountain elf"),
			new Node(db, ontology, 8, "enchanted troll"),
			new Node(db, ontology, 9, "lesser deamon turtle"),
			new Node(db, ontology, 10, "lesser warlord of Ka'a")
		};

		for (Node node : nodes) {
			db.storeNode(node);
		}

		testDictionaryQuery(db, "enchanted", new int[]{3, 8});
		testDictionaryQuery(db, "lesser", new int[]{9, 10});
		testDictionaryQuery(db, "phantom", new int[]{0});
		testDictionaryQuery(db, "warlord", new int[]{0, 10});
		testDictionaryQuery(db, "of", new int[]{});
		testDictionaryQuery(db, "headed", new int[]{5, 6});
	}

	public void testDictionaryAdvanced() {
		System.out.println("testDictionaryAdvanced");

		// Create the test database
		OntologyDatabase db = new OntologyDatabase(getTestDriver());
		OntologyHandle onto = db.getOntology("oiski_poiski");		

		// Store all nodes
		for (NodeDescriptor node : OntologyDatabaseDriverTestData.getTestDescriptors(onto)) {
			db.storeNode(node);
		}

		// Test some dictionary queries
		testDictionaryQuery(db, "bear", new int[]{67, 38, 49});

		// Delete the three nodes
		db.deleteNode(new NodeHandle(onto, 67));
		db.deleteNode(new NodeHandle(onto, 38));
		db.deleteNode(new NodeHandle(onto, 49));

		// We should not find "bear" anymore
		testDictionaryQuery(db, "bear", new int[]{}, true);
	}
}
