/*
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.ontology.driver;

import de.unibi.techfak.scie.ontology.driver.NodeHandle;
import de.unibi.techfak.scie.ontology.driver.OntologyHandle;
import de.unibi.techfak.scie.ontology.driver.AssociationDirection;
import de.unibi.techfak.scie.ontology.driver.NodeDescriptor;
import de.unibi.techfak.scie.ontology.driver.NodeName;
import de.unibi.techfak.scie.ontology.driver.AssociationType;
import de.unibi.techfak.scie.ontology.driver.ComparisonMode;
import de.unibi.techfak.scie.ontology.driver.OntologyDatabaseDriver;
import de.unibi.techfak.scie.ontology.driver.Association;
import java.lang.Thread;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Collection;
import java.util.Date;
import java.util.List;
import java.util.TreeSet;
import java.util.logging.Level;
import java.util.logging.Logger;
import junit.framework.TestCase;
import org.apache.commons.lang3.ArrayUtils;

/**
 *
 * @author andreas
 */
public abstract class OntologyDatabaseDriverTest extends TestCase {

	public OntologyDatabaseDriverTest(String testName) {
		super(testName);
	}

	abstract protected OntologyDatabaseDriver getTestInstance();

	/**
	 * Tests the "getOntology", "getOntologies" and "deleteOntology" methods.
	 */
	public void testOntologies() {
		final String ONTOLOGY_NAME_1 = "test_ontology";
		final String ONTOLOGY_NAME_2 = "Test_ontology";

		System.out.println("Ontology Management");
		OntologyDatabaseDriver instance = getTestInstance();

		// Database should be empty
		assertEquals("No ontologies", 0, instance.getOntologies().size());

		// Get the handle to a new ontology
		OntologyHandle ontology = instance.getOntology(ONTOLOGY_NAME_1);
		assertTrue("Ontology is not null", ontology != null);
		assertEquals("Ontology has correct name", ONTOLOGY_NAME_1, ontology.getName());

		// Database should now have one ontology
		assertEquals("Ontology count", 1, instance.getOntologies().size());

		// The two ontologies must equal
		OntologyHandle ontology2 = instance.getOntology(ONTOLOGY_NAME_1);
		assertEquals("Ontologies equal", ontology, ontology2);

		// The ontology date should be set to some date around the last few seconds
		long now = (new Date()).getTime();
		Date d1 = new Date(now - 5000);
		Date d2 = new Date(now + 1000);
		Date md = instance.getOntologyModificationDate(ontology);
		assertTrue("OntologyDate", md.after(d1));
		assertTrue("OntologyDate", md.before(d2));

		Date d3 = new Date(1234567);
		instance.setOntologyModificationDate(ontology, d3);

		md = instance.getOntologyModificationDate(ontology);
		assertTrue("OntologyDate", Math.abs(d3.getTime() - md.getTime()) < 1000);

		// The two ontologies must not equal, now there should be two ontologies
		OntologyHandle ontology3 = instance.getOntology(ONTOLOGY_NAME_2);
		assertTrue("Ontologies inequal", !ontology.equals(ontology3));
		assertEquals("Ontology count", 2, instance.getOntologies().size());

		// Deleting an ontology must reduce the ontology count and delete the
		// correct one
		instance.deleteOntology(ontology);
		assertEquals("Ontology count", 1, instance.getOntologies().size());
		for (OntologyHandle onto : instance.getOntologies()) {
			assertEquals("Remaining ontology", ONTOLOGY_NAME_2, onto.getName());
		}
	}

	/**
	 * Tests the getRootNodes, getNodes, storeNodes and deleteNodes methods.
	 */
	public void testNodes() {
		System.out.println("Node Management");
		OntologyDatabaseDriver instance = getTestInstance();

		// Insert all test nodes
		OntologyHandle onto = instance.getOntology("oiski_poiski");
		instance.storeNodes(Arrays.asList(OntologyDatabaseDriverTestData.getTestDescriptors(onto)));

		// There must be exactly one root node named "noahs island"
		Collection<NodeDescriptor> rootNodes = instance.getRootNodes(onto);
		assertEquals("Root node count", 1, rootNodes.size());
		for (NodeDescriptor node : rootNodes) {
			assertEquals("Root node name", "noahs island", node.getPrimaryName());
			assertEquals("Root node id", 0, node.getId());
		}

		// Fetch the node content for node 7
		List<NodeDescriptor> node7s = instance.getNodes(Arrays.asList(
				new NodeHandle[]{new NodeHandle(onto, 7)}));
		assertEquals("Node 7 count", 1, node7s.size());
		NodeDescriptor node7 = node7s.get(0);
		assertEquals("Node 7 id", 7, node7.getId());
		assertEquals("Node 7 name:name", "sacha", node7.getPrimaryNameDescr().getName());
		assertEquals("Node 7 name:type", null, node7.getPrimaryNameDescr().getType());

		// Update node 7 -- set name type to "oiski poiski!" and make it a root
		// node
		NodeDescriptor newNode7 = new NodeDescriptor(onto, 7,
				new NodeName("sacha", "oiski poiski!"), true);
		instance.storeNodes(Arrays.asList(new NodeDescriptor[]{newNode7}));

		// Fetch the node content for node 7
		node7s = instance.getNodes(Arrays.asList(
				new NodeHandle[]{new NodeHandle(onto, 7)}));
		assertEquals("Node 7 count", 1, node7s.size());
		node7 = node7s.get(0);
		assertEquals("Node 7 id", 7, node7.getId());
		assertEquals("Node 7 name:name", "sacha", node7.getPrimaryName());
		assertEquals("Node 7 name:type", "oiski poiski!", node7.getPrimaryNameDescr().getType());

		rootNodes = instance.getRootNodes(onto);
		assertEquals("Root node count", 2, rootNodes.size());

		// Delete the original root node
		instance.deleteNode(new NodeHandle(onto, 0));

		rootNodes = instance.getRootNodes(onto);
		assertEquals("Root node count", 1, rootNodes.size());
		for (NodeDescriptor node : rootNodes) {
			assertEquals("Root node name", "sacha", node7.getPrimaryName());
			assertEquals("Root node id", 7, node.getId());
		}

		// Make sure null is returned if the deleted node is queried
		assertEquals("Deleted node is null", null,
				instance.getNode(new NodeHandle(onto, 0)));

		// Delete the new root node as well
		instance.deleteNode(new NodeHandle(onto, 7));

		rootNodes = instance.getRootNodes(onto);
		assertEquals("Root node count", 0, rootNodes.size());

		// Check the existance of some nodes
		List<NodeHandle> nodes = Arrays.asList(new NodeHandle[]{
			new NodeHandle(onto, 0),
			new NodeHandle(onto, 7),
			new NodeHandle(onto, 16)
		});
		List<Boolean> exist = instance.queryNodesExist(nodes);
		assertEquals("Exist count", 3, exist.size());
		assertEquals("Node 0", (Boolean) false, exist.get(0));
		assertEquals("Node 7", (Boolean) false, exist.get(1));
		assertEquals("Node 16", (Boolean) true, exist.get(2));

		// Test querying a list of nodes
		List<NodeDescriptor> nodeDescrs = instance.getNodes(nodes);
		assertEquals("Node descriptors count", 3, nodeDescrs.size());
		assertEquals("Node 0", null, nodeDescrs.get(0));
		assertEquals("Node 7", null, nodeDescrs.get(1));
		assertTrue("Node 16", nodeDescrs.get(2) != null);

		// Check the node 16
		NodeDescriptor node16 = nodeDescrs.get(2);
		assertEquals("Node 16 id", 16, node16.getId());
		assertEquals("Node 16 ontology", onto, node16.getOntology());
		assertEquals("Node 16 name:name", "the valve rhino", node16.getPrimaryName());
		assertEquals("Node 16 name:type", null, node16.getPrimaryNameDescr().getType());

		// Delete the name of node 16
		node16.names.clear();
		instance.storeNode(node16);

		// The node should still by queryable
		List<NodeDescriptor> node16s = instance.getNodes(Arrays.asList(
				new NodeHandle[]{new NodeHandle(onto, 16)}));
		assertEquals("Node 16 count", 1, node16s.size());
		node16 = node16s.get(0);
		assertEquals("Node 16 id", 16, node16.getId());
		assertEquals("Node 16 ontology", onto, node16.getOntology());
		assertEquals("Node 16 name", null, node16.getPrimaryNameDescr());

		// Add three new names to node 16
		NodeName names[] = {
			new NodeName("the valve rhino", "season 01", ComparisonMode.CLEAN),
			new NodeName("the thunder rhino", "season 02", ComparisonMode.FUZZY),
			new NodeName("the giant rhino", "season 03", ComparisonMode.LOWERCASE),
			new NodeName("the superfluous rhino", "season 04", true),
		};
		node16.names.addAll(Arrays.asList(names));
		instance.storeNode(node16);

		// Querying the node should work
		node16 = instance.getNode(new NodeHandle(onto, 16));
		assertTrue("Node 16 exists", node16 != null);
		assertEquals("Node 16 name count", 4, node16.getNames().size());
		for (NodeName name : node16.getNames()) {
			int idx = ArrayUtils.indexOf(names, name);
			assertTrue("Node 16 name", idx >= 0);
			assertTrue("Node 16 name", names[idx] != null);
			assertEquals("Node 16 name", names[idx].getComparisonMode(), name.getComparisonMode());
			names[idx] = null;
		}

		// The primary node name shoud be "the superfluous rhino"
		assertEquals("Node 16 primary name", "the superfluous rhino", node16.getPrimaryName());
	}

	/**
	 * Tests the getAssociations, storeAssociations and deleteAssociations
	 * functions.
	 */
	public void testAssociations() {
		System.out.println("Association Management");
		OntologyDatabaseDriver instance = getTestInstance();

		// Insert all test nodes
		OntologyHandle onto = instance.getOntology("oiski_poiski");
		instance.storeNodes(Arrays.asList(OntologyDatabaseDriverTestData.getTestDescriptors(onto)));

		// Insert all test associationas
		instance.storeAssociations(Arrays.asList(OntologyDatabaseDriverTestData.getTestAssociations(onto)));

		// Check: Sacha IS_A connections (outbound)
		Collection<Association> associations;
		associations = instance.getAssociations(new NodeHandle(onto, 7),
				AssociationType.IS_A.getValue(), AssociationDirection.OUTBOUND);

		// There are two outbound IS_A associations
		assertEquals("Association count", 2, associations.size());
		int targetIds[] = new int[]{3, 37};
		for (Association assoc : associations) {
			assertEquals("Association type", AssociationType.IS_A, assoc.getType());
			assertEquals("Association ontology", onto, assoc.getSource().getOntology());
			assertEquals("Association ontology", onto, assoc.getTarget().getOntology());
			assertEquals("Association source", new NodeHandle(onto, 7), assoc.getSource());
			int idx = ArrayUtils.indexOf(targetIds, assoc.getTarget().getId());
			assertTrue("Association target", idx >= 0);
			assertTrue("Association target", targetIds[idx] >= 0);
			targetIds[idx] = -1;
		}

		// Check: All animals that are primates
		associations = instance.getAssociations(new NodeHandle(onto, 35),
				AssociationType.IS_A.getValue(), AssociationDirection.INBOUND);

		// There are six inbound IS_A associations
		assertEquals("Association count", 6, associations.size());
		int sourceIds[] = new int[]{40, 45, 46, 48, 58, 62};
		for (Association assoc : associations) {
			assertEquals("Association type", AssociationType.IS_A, assoc.getType());
			assertEquals("Association ontology", onto, assoc.getSource().getOntology());
			assertEquals("Association ontology", onto, assoc.getTarget().getOntology());
			assertEquals("Association target", new NodeHandle(onto, 35), assoc.getTarget());
			int idx = ArrayUtils.indexOf(sourceIds, assoc.getSource().getId());
			assertTrue("Association source", idx >= 0);
			assertTrue("Association source", sourceIds[idx] >= 0);
			sourceIds[idx] = -1;
		}

		// Check: Both connections, inbound and outbound (test node: "nohas island")
		associations = instance.getAssociations(new NodeHandle(onto, 4),
				AssociationType.CONTAINS_A.getValue(), AssociationDirection.BOTH);

		// There are 22 CONTAINS_A associations (in this case all are targets!)
		assertEquals("Association count", 22, associations.size());
		targetIds = new int[]{6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17,
			18, 19, 20, 21, 22, 23, 24, 25, 26, 44};
		for (Association assoc : associations) {
			assertEquals("Association type", AssociationType.CONTAINS_A, assoc.getType());
			assertEquals("Association ontology", onto, assoc.getSource().getOntology());
			assertEquals("Association ontology", onto, assoc.getTarget().getOntology());
			assertEquals("Association source", new NodeHandle(onto, 4), assoc.getSource());
			int idx = ArrayUtils.indexOf(targetIds, assoc.getTarget().getId());
			assertTrue("Association target", idx >= 0);
			assertTrue("Association target", targetIds[idx] >= 0);
			targetIds[idx] = -1;
		}

		// Check: Both connections, inbound and outbound (test node: "nohas island")
		associations = instance.getAssociations(new NodeHandle(onto, 4),
				AssociationType.IS_A.getValue(), AssociationDirection.BOTH);

		// There are 1 IS_A associations (in this case all are targets!)
		assertEquals("Association count", 1, associations.size());
		targetIds = new int[]{3};
		for (Association assoc : associations) {
			assertEquals("Association type", AssociationType.IS_A, assoc.getType());
			assertEquals("Association ontology", onto, assoc.getSource().getOntology());
			assertEquals("Association ontology", onto, assoc.getTarget().getOntology());
			assertEquals("Association source", new NodeHandle(onto, 4), assoc.getSource());
			int idx = ArrayUtils.indexOf(targetIds, assoc.getTarget().getId());
			assertTrue("Association target", idx >= 0);
			assertTrue("Association target", targetIds[idx] >= 0);
			targetIds[idx] = -1;
		}

		// So if we query for all associations there should be 23
		associations = instance.getAssociations(new NodeHandle(onto, 4),
				AssociationType.ALL, AssociationDirection.BOTH);
		assertEquals("Association count", 23, associations.size());
		targetIds = new int[]{3, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17,
			18, 19, 20, 21, 22, 23, 24, 25, 26, 44};
		for (Association assoc : associations) {
			assertEquals("Association ontology", onto, assoc.getSource().getOntology());
			assertEquals("Association ontology", onto, assoc.getTarget().getOntology());
			assertEquals("Association source", new NodeHandle(onto, 4), assoc.getSource());
			int idx = ArrayUtils.indexOf(targetIds, assoc.getTarget().getId());
			assertTrue("Association target", idx >= 0);
			assertTrue("Association target", targetIds[idx] >= 0);
			targetIds[idx] = -1;
		}

		// Add an association between 3 to 4
		instance.storeAssociation(
				new Association(new NodeHandle(onto, 3), new NodeHandle(onto, 4),
						AssociationType.CONTAINS_A));

		// This should increase the association count for node 4 to 24
		associations = instance.getAssociations(new NodeHandle(onto, 4),
				AssociationType.ALL, AssociationDirection.BOTH);
		assertEquals("Association count", 24, associations.size());

		// Inserting the same association with another type should not further
		// increment the association count
		instance.storeAssociation(
				new Association(new NodeHandle(onto, 3), new NodeHandle(onto, 4),
						AssociationType.DEFAULT));
		associations = instance.getAssociations(new NodeHandle(onto, 4),
				AssociationType.ALL, AssociationDirection.BOTH);
		assertEquals("Association count", 24, associations.size());

		// Deleting the association should decrement the count by one, independent
		// of the association type
		instance.deleteAssociation(
				new Association(new NodeHandle(onto, 3), new NodeHandle(onto, 4),
						AssociationType.SYNONYM_OF));
		associations = instance.getAssociations(new NodeHandle(onto, 4),
				AssociationType.ALL, AssociationDirection.BOTH);
		assertEquals("Association count", 23, associations.size());

		// Inserting an association between non-existing nodes should create
		// these.
		assertTrue("Not existing nodes", !instance.queryNodeExists(new NodeHandle(onto, 1000)));
		assertTrue("Not existing nodes", !instance.queryNodeExists(new NodeHandle(onto, 1001)));
		instance.storeAssociation(new Association(new NodeHandle(onto, 1000), new NodeHandle(onto, 1001)));
		assertTrue("Now existing nodes", instance.queryNodeExists(new NodeHandle(onto, 1000)));
		assertTrue("Now existing nodes", instance.queryNodeExists(new NodeHandle(onto, 1001)));

		// The created nodes should not be root nodes
		assertTrue("Not root node", !instance.getNode(new NodeHandle(onto, 1000)).isRoot());
		assertTrue("Not root node", !instance.getNode(new NodeHandle(onto, 1001)).isRoot());
	}

	/**
	 * Test the queryDictionary, storeDictionaryEntries and
	 * deleteDictionaryEntries methods.
	 */
	public void testDictionary() {
		System.out.println("Dictionary Management");
		OntologyDatabaseDriver instance = getTestInstance();

		// Insert all test nodes
		OntologyHandle onto = instance.getOntology("oiski_poiski");
		instance.storeNodes(Arrays.asList(OntologyDatabaseDriverTestData.getTestDescriptors(onto)));

		// Insert some test dictionary entries
		String names1[] = new String[]{"sacha", "russian", "desman"};
		String names2[] = new String[]{"noah", "polar", "bear"};
		instance.storeDictionaryEntry(new NodeHandle(onto, 7), new TreeSet(Arrays.asList(names1)));
		instance.storeDictionaryEntry(new NodeHandle(onto, 37), new TreeSet(Arrays.asList(names1)));
		instance.storeDictionaryEntry(new NodeHandle(onto, 6), new TreeSet(Arrays.asList(names2)));

		// Query "russian"
		Collection<NodeDescriptor> nodes = instance.queryDictionary("russian");

		// There should be two nodes associated to that
		assertEquals("Node count", 2, nodes.size());
		int ids[] = new int[]{7, 37};
		String nodeNames[] = new String[]{"sacha", "russian desman"};
		for (NodeDescriptor node : nodes) {
			int idx = ArrayUtils.indexOf(ids, node.getId());
			assertEquals("Node ontology", onto, node.getOntology());
			assertTrue("Node id", idx >= 0);
			assertTrue("Node id", ids[idx] != -1);
			ids[idx] = -1;

			idx = ArrayUtils.indexOf(nodeNames, node.getPrimaryName());
			assertTrue("Node name", idx >= 0);
			assertTrue("Node name", nodeNames[idx] != null);
			nodeNames[idx] = null;
		}

		// Deleting the dictionary entries for node 7 and repeat the test
		instance.deleteDictionaryEntry(new NodeHandle(onto, 7));

		// Query "russian"
		nodes = instance.queryDictionary("russian");

		// There should be one node associated to that
		assertEquals("Node count", 1, nodes.size());
		ids = new int[]{37};
		nodeNames = new String[]{"russian desman"};
		for (NodeDescriptor node : nodes) {
			int idx = ArrayUtils.indexOf(ids, node.getId());
			assertEquals("Node ontology", onto, node.getOntology());
			assertTrue("Node id", idx >= 0);
			assertTrue("Node id", ids[idx] != -1);
			ids[idx] = -1;

			idx = ArrayUtils.indexOf(nodeNames, node.getPrimaryName());
			assertTrue("Node name", idx >= 0);
			assertTrue("Node name", nodeNames[idx] != null);
			nodeNames[idx] = null;
		}
	}

	/**
	 * Test the consistency of the database (cross-tests many methods).
	 */
	public void testConsistency() {
		System.out.println("Consistency Test");
		OntologyDatabaseDriver instance = getTestInstance();

		// Insert all test nodes and all associations plus a few dictionary
		// entries
		String names1[] = new String[]{"sacha", "russian", "desman"};
		String names2[] = new String[]{"noah", "polar", "bear"};
		OntologyHandle onto = instance.getOntology("oiski_poiski");
		instance.storeNodes(Arrays.asList(OntologyDatabaseDriverTestData.getTestDescriptors(onto)));
		instance.storeAssociations(Arrays.asList(OntologyDatabaseDriverTestData.getTestAssociations(onto)));
		instance.storeDictionaryEntry(new NodeHandle(onto, 7), new TreeSet(Arrays.asList(names1)));
		instance.storeDictionaryEntry(new NodeHandle(onto, 37), new TreeSet(Arrays.asList(names1)));
		instance.storeDictionaryEntry(new NodeHandle(onto, 6), new TreeSet(Arrays.asList(names2)));

		// Delete node 7
		instance.deleteNode(new NodeHandle(onto, 7));
		
		// Make sure: The node does not exist
		assertEquals("Node has vanished", false, instance.queryNodeExists(new NodeHandle(onto, 7)));

		// Make sure: No associations from node 7 to anywhere
		Collection<Association> assocs = instance.getAssociations(new NodeHandle(onto, 7), AssociationType.ALL,
				AssociationDirection.BOTH);
		assertEquals("No associations", 0, assocs.size());

		// Make sure: No associations left from node 3 to 7
		assocs = instance.getAssociations(new NodeHandle(onto, 3), AssociationType.ALL,
				AssociationDirection.BOTH);
		for (Association assoc: assocs) {
			assertTrue("No associations", assoc.getSource().getId() != 7);
			assertTrue("No associations", assoc.getTarget().getId() != 7);
		}

		// Make sure: No associations left from node 37 to 7
		assocs = instance.getAssociations(new NodeHandle(onto, 37), AssociationType.ALL,
				AssociationDirection.BOTH);
		for (Association assoc: assocs) {
			assertTrue("No associations", assoc.getSource().getId() != 7);
			assertTrue("No associations", assoc.getTarget().getId() != 7);
		}

		// Make sure: No dictionary entries left for node 7
		// Query "russian"
		Collection<NodeDescriptor> nodes = instance.queryDictionary("russian");

		// There should be one node associated to that
		assertEquals("Node count", 1, nodes.size());
		int ids[] = new int[]{37};
		String nodeNames[] = new String[]{"russian desman"};
		for (NodeDescriptor node : nodes) {
			int idx = ArrayUtils.indexOf(ids, node.getId());
			assertEquals("Node ontology", onto, node.getOntology());
			assertTrue("Node id", idx >= 0);
			assertTrue("Node id", ids[idx] != -1);
			ids[idx] = -1;

			idx = ArrayUtils.indexOf(nodeNames, node.getPrimaryName());
			assertTrue("Node name", idx >= 0);
			assertTrue("Node name", nodeNames[idx] != null);
			nodeNames[idx] = null;
		}
	}

}
